{ MineSweeper solver.
  This program solves the MineSweeper game.
  After the user indicated the program which is the MineSweeper Window,
  the program moves alone the mouse cursor to click on cells and
  solve the game.
  In  tEx063.pp  a simple strategie was used, here the strategie will be
  more sophisticated, using probabilities to decide on which cell to click.

  Here are some rules.
  """"""""""""""""""""
  Simple Rule one :
  A cell (nX,nY) has no bomb for sure, if a neighbour cell indicate a number equal
  to the number of its neighbour having a Flag. i.e.
  if aabBO[i,j] = 0   for (i,j) a neighbour of (nX,nY)
  It is simpler to say that if a cell indicate as many bombs as it has
  neighbour with a flag, then all other neighbour has no bombs.
  But the other way has the advantage of taking only decisions about
  unvisited cells.

  Simple (nX,nY) Rule two :
  A cell has a bomb for sure, if a neighbour cell idicate as many bombs as
  the number of unvisited cells minus those having a Flag. i.e.
  if aabBO[i,j] = aabTS[i,j]  for (i,j) a neighbour of (nX,nY)
  It is simpler to say that if a cell indicate as many bombs as it has
  neighbour with a flag plus unvisited neighbours, then all
  unvisited neighbour have a bombs.

  Many other special rules can be stated. But they will be explored
  systematically, another way.
  Here are some drawings giving examples of other rules.
  x = a visited cell
  a digit = number of neighbour cells having a bomb - those already Flagged
  . = a cell, of any type
  o = has no bombs for sure after deduction

  x . . o    x . . o     x x . o o     x x x
  x 1 1 o    . 2 1 o     x 1 . 1 o     x 1 . o    etc.
  x . . o    x . . o     x x . o o     x . 1 o
                                         o o o

  If no cell can be found without a bomb, using the simple rules, then
  for each cell, the program will see what are the deductions if a bomb is
  put into the cell. If it leads to a contradiction, then for sure this
  cell has no bombs.
  The contradiction will be that  aabBO[i,j] < 0  for a cell (i,j), i.e.
  more flags has to be set than there are free neighbour cells.
  --------------------------------------------}
program tEx065;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses

  wApplication0
 ,wWindowText0
// ,wWindow0
 ,wDialog0
 ,wWindowBase0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wErrors0
 ,wVariableNumLink2
 ,wMessages2
 ,wGlobals2
 ,wDebug0
 ,Strings
 ,Windows;

//const
//  // To compute the optimal cell to click on, to win
//  MaxConfigStore = 10000; // Maximum number of configurations stored
//  MaxFreeCellStore = 60; // Maximum number of cell stored
//
//type
//  TypeConfigStore = record
//  //""""""""""""""""""""""
//  anC      // Content of the cell.
//          : array[1..MaxFreeCellStore] of integer;
//    // The cells with bombs are not stored, only thoses without bombs.
//  end;

type
  Pinteger = ^integer;
  TMineSolver = object
  //""""""""""""""""""
  poWinMain     : PWindowText;
  // Handle of the Main window

  hWinSec       : HWND;
  // Handle of the other window

  colGray : TypeColor;
  // Gray color of the cels.

  nSpeed : integer;
  // To solve faster and not see the mouse mouvments, or slower
  // 0 is the slowest, very slow, 10 wait per mouse mouvements
  // 1 is less slow,  8 wait per mouse mouvements
  // 2 = 5 wait, 3=3 wait, 4=1 wait
  // 5 = wait only 3/4 of the mouse mouvement, 6=wait 1/2 of mouv.
  // 10 = the fastest possible.

  fIsKeyUp    : boolean;
  // =TRUE is the  F or S  key has been pressed,
  // In order to not get twice the instruction to change the mouse speed

  ptTopLeft : POINT;
  // Position in pixels in the client coordinate of the top left corner
  // of the top left mine cell.

  ptTopLeftScreen : POINT;
  // Position in pixels in the Screen coordinate of the top left corner
  // of the top left mine cell.

  ptRightBottomMargin : POINT;
  // Size of the right and Bottom margin of the Minesweeper game.
  // Used to compute the number of cells in the game.

  nXDelta, nYDelta : integer;
  // number of pixels between two mine cells.

  nXNumber, nYNumber : integer;
  // number of mine cells on a line and on a column

  fQuitSolve : boolean;
  // Set to TRUE to quit the solver

  nNbBombExplosed : integer;
  // Number of bombs which exploded.
  // If more than zero, we lost.

  abXFree,abYFree : array[0..2500] of byte;
  // List of all cells which have not been visited.

  nNbFreeCells : integer;
  // Number of not visited cells.

  nNbBombsTot : integer;
  // Total number of bombs. This must be given by the user,
  // in the custom mode.

  nNbBombs : integer;
  // number of unresolved bombs, i.e.
  // number of cells having a bomb, but no flag on it.

  nNbGames : int32;
  // number of played games

  aanBObis, // copy
  aanBO : array[0..36,0..31] of integer;
  // The content of the MinerSweep
  // Row and column 0 are used to simplify computation, but
  // First mine cell Row and Column begins with index 1.
  // if aanBO[i,j] < 9, then it contains the number of neighbour not visited
  // (i.e. marked cteToSee) which have a bomb.

  aanTSbis, // copy
  aanTS : array[0..36,0..31] of integer;
  // number of neighbour cels Too See, i.e.
  // not having been explored.

  // To do backtracking, for searching cells having low probability of
  // having a bomb.
  apnModif : array[0..2500] of Pinteger;
  // list of modified variables which has to be set back to their original value

  anModifValue : array[0..2500] of integer;
  // List of the values which have been modifies.

  nIndModif : integer;
  // Index of the two previous arrays.
  // Above it, indexes have not been defined.

  nNbFreeCellsBis : integer;
  // during backtracking, number of unvisited cells.

  nNbBombsBis : integer;
  // during backtracking, number of left bombs.

  nNbFreeCellsInfo : integer;
  // Number of free cells having neighbour containing information.
  // Set by sorting, such that
  // (abXFree,abYFree)[1..nNbFreeCellsInfo] are those interresting
  // To test if they contain bombs or not

  avNbWithFlag : array[1..2500] of Float;
  // For each cell, number of configuration found with the cell having a bomb
  // Used in  ProbabilityExplore  and  ProbabilityExploreMore

  vPrCellWIWithBomb : Float;
  // To compute the probability that a cell without info has a bomb.

  vTotalConfig : Float;
  // Total number of possible configurations found
  // Used in  ProbabilityExplore  and  ProbabilityExploreMore

  vProbProduct : Float;
  // product of (1-prob)  where prob was the probability that the cell
  // contain a bomb. In most cases, prob = 0, but at starting and
  // sometimes, decisions must be made with a certain probability.

  nMethodContent : integer;
  // Method used to recognise the content of the cells

  nGameVersion : integer;
  // Select the game version.
  // 951 => the windows 95 minesweeper game
  // 981 => the windows 98 minesweeper game
  // 982 => the windows 98 minesweeper game with another cells recognition
  //  0 => another minesweeper game

  anXPosCol, anYPosCol : array[1..8] of integer;
  // Pixels to selct a symbol by color

  anSymbole : array[0..14] of integer;
  // List of all symboles

  ancolSymbole : array[0..14,1..8] of TypeColor;
  // ancolSymbole[nS,nP] = Colors of the symbole anSymbole[nS]
  // at a pixel position  (anXPosCol[nP],anYPosCol[nP])

  nExitSolver : integer;
  // Can ask to stop the solver on certain conditions.
  //   0 => do not stop
  //   1 => Stop if no cells without bombs found easilly
  // 101 => Stop if no cells without bombs found easilly
  //        but cells found by deduction
  //   2 => Stop if no cells without bombs found by deduction
  // 102 => Stop if no cells without bombs found by deduction
  //        but cells found to be without bombs by probabilistic search
  //   3 => Stop if no cells found without bombs by the 100% sure detection methode

  // To show the value of variables in a dialog box
  nXPosCell, nYPosCell : int32;
  vBombProbability     : Float;
  // Probability that the cell (nXPosCell, nYPosCell) has a bomb.

  // Windows in which information will be displayed :
  //"""""""""""""""""""""""""""""""""""""""""""""""""
  nWinSeeMineContent : integer;

  // For error display in GetMineContentCol
  nWinGetMinCont1,
  nWinGetMinCont2 : integer;

  // in ProbabilityExploreMore to display where has been clicked
  // with probability do boom
  nWinProbabilityExploreMore1 : integer;

  // To display statistics :
  nWinGetStatistiquePlay1,
  nWinGetStatistiquePlay2 : integer;

  nDisplayFrequ : integer;
  // Display how many games have been played every  nDisplayFrequ  games.

  // Debugging variables
  //""""""""""""""""""""
  fShowCellsOrgPixels : boolean;
  // During cells recognition, indicate the (0,0) pixel of each cell
{
  // To get statistics.
  fHadFewConfig : boolean;
  nNbFewConfigWin,
  nNbFewConfigLost  : int32;

  //--------------------------------------------------------
  // To Compute the optimal cell where to click, from the complete
  // list of possible configurations.
  // Quit difficulte and time consuming, for very little results.
  arConfig : array[0..MaxConfigStore] of TypeConfigStore;
  // List of all possible configurations compatible with the actual state
  // of the minesweeper

  nNbConfig : int32;
  // Number of configurations stored in arConfig
}
  constructor Init;
  destructor Done; virtual;

  procedure SetColorsWin98;
  // Colors for the windows 98 minesweeper symboles, to
  // select them by colors

  procedure SetColorsWin95;
  // Colors for the windows 95 minesweeper symboles, to
  // select them by colors

  procedure SelectGameVersion;
  // Select the version of minesweeper.
  // Actually, two version are programmed, Win95 and Win98.

  function IsSecWindowDefined:boolean;
  // if hWinSec <> 0  then return TRUE
  // else display a dialog box explaining what to do, and return FALSE

  procedure MarkCell(nXP,nYP:integer; wColorP:word32);
  // Put a mark on the bottom right of the cell.

  procedure GetSecWindowPosition;
  procedure SimulateLeftClick;
  procedure SimulateRightClick;

  procedure MoveMouseTo(nX2,nY2:integer);
  // Moving the mouse cursor to the cell nX2, nY2.

  procedure ClickOnNewGame;
  // Click on the new game icone to begin a new minesweeper game

  procedure InitCellsData;
  // Initialisation to set that no cells have been visited, to see their content,
  // and other array initialisation

  procedure GetCellPositions;

  function GetMineContentCol(hDCS:HDC; nXP,nYP:integer):integer;
  // For the Windows 98 minesweeper game
  // Get using colors

  function GetMineContentWin98(hDCS:HDC; nXP,nYP:integer):integer;
  // For the Windows 98 minesweeper game
  // nXP, nYP = coordinate of a mine cell.
  // return the content of it.
  // = cteToSee if it has not been explored, 0 if empty,
  // = cteQuestion if a question mark is in
  // = cteFlag  if a flag is in
  // = cteBoom  if a bomb is in, and thus the game finished.

  procedure GetMineSweeperContent(wFlagsP:word32);
  // Get the content of the complete minesweeper window
  // wFlagsP = 1 => show the content in the debug window.
  // wFlagsP = 6 => Cells with Flags and Question are set to be not visited

  procedure UpdateMineSweeperContent;
  // The content of the mineSweeper is supposed to have been set right
  // before the last click, from it, update the information about
  // the content.

  function PutFlags(wFlagsP:word32):integer;
  // Put Flags on cells where it is sure that a bomb existe.
  // Return the number of flag added.
  // wFlagsP = 0 => normal playing.
  // wFlagsP = 1 => do not play, only count the number of cells which could
  // are detected having a bombs and not being flagged.

  procedure StartingExplore;
  // When we beging, start with some special cells.

  function SaftyExplore(wFlagsP:word32):integer;
  // Click on cells which have no bombs for sure
  // Return the number of cells which have been clicked.
  // wFlagsP = 0 => normal playing.
  // wFlagsP = 1 => do not play, only the number of cells which could
  // are detected having no bombs.

  procedure SaveValue(pnValue:Pinteger);
  // Save in  apnModif[nIndModif]  and anModifValue[nIndModif]
  // the pointer  pnValue,  and the content of pnValue^

  function SetFlagTo(nXP,nYP:integer):integer;
  // Try to set a  cteFlag  or  cteNoBomb to a cell, to see the
  // logical consequeces, and find contradictions. Thus finding cells
  // having no bombs, or having bombs for sure.

  function PutFlagsMore(wFlagsP:word32):integer;
  // Put Flags on cells where it is sure that a bomb existe.
  // Use more deduction than in PutFlags to find these cells
  // Return the number of flag added.
  // wFlagsP = 0 => normal playing.
  // wFlagsP = 1 => do not play, only count the number of cells which could
  // are detected having a bombs and not being flagged.

  function SaftyExploreMore(wFlagsP:word32):integer;
  // Click on cells which have no bombs for sure
  // But the searching of these cells is more sophisticated than SaftyExplore
  // wFlagsP = 0 => normal playing.
  // wFlagsP = 1 => do not play, only count the number of cells which could
  // are detected having no bombs.

  procedure SortFreeCells;
  // Sort the abXFree,abYFree arrays such that cells having few unvisited
  // neighbour come first.
{
  procedure ConfigStore(nNbConfigWI:integer);
  // Store in  arConfig  the configuration compatible with the actual state
  // of the game.
  // nNbConfigWI = number of configurations of the cells Without Info

  function NbLostConfig(
       nCell_0:integer; // cell on which we click
   var anConfList:array of integer; // list of possible configurations
       nNbConfList:integer; // number of possible configurations
   var anCellFree:array of integer; // list of unvisited cells of the configurations
       nNbCellFree:integer; // number of unvisited cells
       nNbLostConfigMax:integer // maximum loosing configurations accepted.
   ):integer;
  // return the number of loosing configurations if the cell "nCell_0"
  // is clicked on.
  // Use NbWinConfig recursively to compute its value.
  // Can be very time consuming
  // We know a cell where clicking on it will give at most
  // nNbLostGameMax loosing config. Thus it is useless to test cells with more
  // than this number of loosing configurations

  function FindBestCell:integer;
  // When every cell has a chance to have a bomb, try to find
  // the cell which maximise the probability to win the game.
  // It is not always the cell which minimise the probability to have a bomb.
  // Return the cell maximising the change to win
  // Cell = (abXFree[return], abYFree[return])
}
  function ProbabilityExplore(wFlagsP:word32):integer;
  // Explore the cells to find which has the lower probabilities of
  // having a bomb. Take into account the number of bombs.
  // In some case, this probability will be zero, and thus
  // it is sure that a cell without bomb will have been found.

  function ProbabilityExploreMore(wFlagsP:word32):integer;
  // No cell without bomb can be found with certitude, thus
  // one cell has to be clicked on, with some chances that a bomb is in it.
  // Return 1 normally.
  // Return 0 if no cells have been clicked. Used in tests.
  // wFlagsP = 0 => normal playing.
  // wFlagsP = 2 => mark the cell where the program would click.

  procedure SolveMineSweeper;
  // Go and click to solve the minesweeper

  procedure GetStatistiquePlay;
  // Plays many Minesweeper games to get statistics on it.

  procedure HintAreFlagCorrect;
  // Determine if the flags are set correctly

  procedure HintHasBomb;
  // Count the number of cells which can be have a bomb and are not
  // already marked

  procedure HintEasillyFound;
  // Count the number of cells which can be easilly found not having a bomb.

  procedure HintDeductionFound;
  // Count the number of cells which can be found by deduction not having a bomb.

  procedure HintProbabilityFound;
  // Count the number of cells which can be found having no bomb,
  // using probabilities. This finds exactly all cells having no bombs for sure.
  // It computes also the probability of having a bomb for each cell.

  procedure HintMarksHasBombDeductionFound;
  // Mark the cells which can be found having a bomb.

  procedure HintMarksHasBombEasyFound;
  // Mark the cells which can be easilly found having a bomb.

  procedure HintMarkEasillyFound;
  // Mark the cells which can be easilly found not having a bomb.

  procedure HintMarkDeductionFound;
  // Mark the cells which can be found having no bomb, by deduction.

  procedure HintMarkProbabilityFound;
  // Mark the cells which can be found having no bomb
  // using probabilities.

  procedure HintMarkProbabilityTryFound;
  // Mark the cells which would be played by the program
  // using probabilities.

  end; // TMineSolver = object

  PMyWindowText = ^TMyWindowText;
  TMyWindowText = object(TWindowText)
  //"""""""""""""""""""""""""""""""""
  constructor Init(strWinTitlePrm:ShortString);
//  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // Create the window.
  end; // TMyWindowText = object(TWindowText)

  PMyWindow = ^TMyWindow;
  TMyWindow = object(TMyWindowText)
  //"""""""""""""""""""""""""""""""
  oSolve : TMineSolver;
  // Minesweeper solver.

  pDialogActivate : PDialog;

  apoDialogHelp : array[0..99] of PDialog;
  // To remember the opened dialog boxes.

  poDialogProb : PDialog;
  // pointer on the dialog indicating the probabilities.

  constructor Init(strWinTitlePrm:ShortString);
  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // Create the window.

  procedure GetSecWindowHandle;
  // To get the handle of the minesweeper window

  procedure ButtonClicked(poCtrl:PWindowBase0; wCtrlID:Word32); virtual;
  // To react when clicking on the button in the dialog 21199

  procedure WMTimer(wPar:word32);
  // handles the timer messages.

  procedure AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:word32; var wRep:word32); virtual;
  // To set to NIL the pointer on the dialog box

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TMyWindowText)

const
  cteFlag     = 20;  // A cell with a bomb has been marked.
  cteQuestion = 30;  // A question mark
  cteNoBomb   = 40;  // A cell without a bomb, but not explored
  cteToSee    = 50;  // unexplored cells
  cteBoom     = 99;  // A bomb has been found

const
  TIMER_ID = 11;

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

// var nNotUsed : integer; // only to avoid an error message while compliling in DevPascal 1.9

constructor TMineSolver.Init;
//===========================
begin
Int32Field(  1001, ptTopLeft.x);
Int32Field(  1002, ptTopLeft.y);
Int32Field(  1003, ptRightBottomMargin.x);
Int32Field(  1004, ptRightBottomMargin.y);
IntegerField(1005, nXDelta);
IntegerField(1006, nYDelta);
IntegerField(1007, nXNumber);
IntegerField(1008, nYNumber);
IntegerField(1009, nSpeed);
IntegerField(1010, nNbBombsTot);
IntegerField(1011, nMethodContent);
IntegerField(1012, nGameVersion);
IntegerField(1013, nExitSolver);

FloatField(  1020, vTotalConfig);
IntegerField(1021, nNbFreeCells);
IntegerField(1022, nNbFreeCellsInfo);
Int32Field(  1051, nXPosCell);
Int32Field(  1052, nYPosCell);
FloatField(  1053, vBombProbability);

IntegerField(1101, nWinSeeMineContent);
IntegerField(1111, nWinGetMinCont1);
IntegerField(1112, nWinGetMinCont2);
IntegerField(1121, nWinProbabilityExploreMore1);
IntegerField(1131, nWinGetStatistiquePlay1);
IntegerField(1132, nDisplayFrequ );
IntegerField(1133, nWinGetStatistiquePlay2);

// Debugging variables
BooleanField(2001, fShowCellsOrgPixels);

poWinMain:=NIL;
hWinSec:=0;
nSpeed :=4; // Slow
fIsKeyUp:=TRUE;

// Usually these values are fixed.
// They can be changed by the user, or computed by the program
ptTopLeft.x  :=12;
ptTopLeft.y  :=55;
ptRightBottomMargin.x:= 8; // 2 pixels less, as a security
ptRightBottomMargin.y:= 8;
nXDelta  :=16;
nYDelta  :=16;
nXNumber :=8;  // This can change, depending on the size of the window
nYNumber :=8;  // This can change, depending on the size of the window

nExitSolver:=0;

// Selection by colors
nMethodContent:=1;

nGameVersion:=981;

SetColorsWin98;

vTotalConfig:=1; // just to be initialised
nXPosCell :=0;
nYPosCell :=0;
vBombProbability:=0.0;

fShowCellsOrgPixels:=FALSE;

// Windows in which information is displayed.
nWinSeeMineContent:=10;

// For error display in GetMineContentCol
nWinGetMinCont1:=0;
nWinGetMinCont2:=0;

nWinProbabilityExploreMore1:=-1;

nWinGetStatistiquePlay1:=2;
nWinGetStatistiquePlay2:=0;
nDisplayFrequ :=100;
end;  // TMineSolver.Init

destructor TMineSolver.Done;
//=========================
begin
end;  // TMineSolver.Done

procedure TMineSolver.SetColorsWin98;
//===================================
// Colors for the windows 98 minesweeper symboles, to
// select them by colors
var
  nS,nP       : integer;
begin
// Define also the margins and the cell size
ptTopLeft.x  :=12;
ptTopLeft.y  :=55;
ptRightBottomMargin.x:= 8; // 2 pixels less, as a security
ptRightBottomMargin.y:= 8;
nXDelta  :=16;
nYDelta  :=16;
nXNumber :=8;  // This can change, depending on the size of the window
nYNumber :=8;  // This can change, depending on the size of the window

nMethodContent:=1; // use color selection to get the game content

for nS:=0 to 14 do begin
   for nP:=1 to 8 do ancolSymbole[nS,nP]:=RGB(192,192,192);

   anSymbole[nS]:=nS;
   end;

// Just to initialise it.
for nP:=1 to 8 do begin
   anXPosCol[nP]:= 1;
   anYPosCol[nP]:= 1;
   end;

// Special symbols
anSymbole[ 9]:=cteFlag;
anSymbole[10]:=cteQuestion;
anSymbole[11]:=cteNoBomb;
anSymbole[12]:=cteToSee;
anSymbole[13]:=cteBoom;  // normal bomb
anSymbole[14]:=cteBoom;  // Red bomb

// First pixel common to many symbols, and color selective
anXPosCol[1]:= 9;
anYPosCol[1]:=12;

ancolSymbole[ 0, 1]:=RGB(192,192,192); // not changed
ancolSymbole[ 1, 1]:=RGB(  0,  0,255);
ancolSymbole[ 2, 1]:=RGB(  0,128,  0);
ancolSymbole[ 3, 1]:=RGB(255,  0,  0);
ancolSymbole[ 4, 1]:=RGB(  0,  0,128);
ancolSymbole[ 5, 1]:=RGB(128,  0,  0);
ancolSymbole[ 6, 1]:=RGB(  0,128,128);
ancolSymbole[ 7, 1]:=RGB(  0,  0,  0);
ancolSymbole[ 8, 1]:=RGB(128,128,128);
ancolSymbole[ 9, 1]:=RGB(  0,  0,  0); // Flag
ancolSymbole[10, 1]:=RGB(192,192,192); // ?
ancolSymbole[11, 1]:=RGB( 17, 19, 31); // Never selected.
ancolSymbole[12, 1]:=RGB(192,192,192); // to see
ancolSymbole[13, 1]:=RGB(  0,  0,  0);
ancolSymbole[14, 1]:=RGB(  0,  0,  0);

// Second pixel color selection, many are useless
anXPosCol[2]:= 1;
anYPosCol[2]:= 1;

ancolSymbole[ 0, 2]:=RGB(192,192,192); // not changed
ancolSymbole[ 7, 2]:=RGB(192,192,192); // not changed
ancolSymbole[ 9, 2]:=RGB(255,255,255); // Flag
ancolSymbole[10, 2]:=RGB(255,255,255); // ?
ancolSymbole[12, 2]:=RGB(255,255,255); // to see
ancolSymbole[13, 2]:=RGB(192,192,192); // black bomb
ancolSymbole[14, 2]:=RGB(255,  0,  0); // bomb on red

// Third pixel color selection, many are useless
anXPosCol[3]:= 8;
anYPosCol[3]:= 4;

ancolSymbole[ 0, 3]:=RGB(192,192,192); // not changed
ancolSymbole[ 7, 3]:=RGB(  0,  0,  0);
ancolSymbole[ 9, 3]:=RGB(255,  0,  0); // Flag
ancolSymbole[10, 3]:=RGB(192,192,192); // ?
ancolSymbole[12, 3]:=RGB(192,192,192); // to see
ancolSymbole[13, 3]:=RGB(  0,  0,  0); // black bomb
ancolSymbole[14, 3]:=RGB(  0,  0,  0); // bomb on red

// Last pixel color selection, many are useless
anXPosCol[4]:= 6;
anYPosCol[4]:= 5;
ancolSymbole[ 7, 4]:=RGB(192,192,192); // 7
ancolSymbole[10, 4]:=RGB(  0,  0,  0); // ?
ancolSymbole[13, 4]:=RGB(  0,  0,  0); // black bomb

// Should be enough
end;  // TMineSolver.SetColorsWin98

procedure TMineSolver.SetColorsWin95;
//===================================
// Colors for the windows 95 minesweeper symboles, to
// select them by colors
var
  nS,nP       : integer;
begin
// Define also the margins and the cell size
ptTopLeft.x  :=11;
ptTopLeft.y  :=54;
ptRightBottomMargin.x:= 11; // 2 pixels less, as a security
ptRightBottomMargin.y:= 11;
nXDelta  :=16;
nYDelta  :=16;
nXNumber :=8;  // This can change, depending on the size of the window
nYNumber :=8;  // This can change, depending on the size of the window

nMethodContent:=1; // use color selection to get the game content

for nS:=0 to 14 do begin
   for nP:=1 to 8 do ancolSymbole[nS,nP]:=RGB(192,192,192);

   anSymbole[nS]:=nS;
   end;

// Just to initialise it.
for nP:=1 to 8 do begin
   anXPosCol[nP]:= 1;
   anYPosCol[nP]:= 1;
   end;

// Special symbols
anSymbole[ 9]:=cteFlag;
anSymbole[10]:=cteQuestion;
anSymbole[11]:=cteNoBomb;
anSymbole[12]:=cteToSee;
anSymbole[13]:=cteBoom;  // normal bomb
anSymbole[14]:=cteBoom;  // Red bomb

// First pixel common to many symbols, and color selective
anXPosCol[1]:=10;
anYPosCol[1]:=12;

ancolSymbole[ 0, 1]:=RGB(192,192,192); // not changed
ancolSymbole[ 1, 1]:=RGB(  0,  0,255);
ancolSymbole[ 2, 1]:=RGB(  0,128,  0);
ancolSymbole[ 3, 1]:=RGB(255,  0,  0);
ancolSymbole[ 4, 1]:=RGB(  0,  0,128);
ancolSymbole[ 5, 1]:=RGB(128,  0,  0);
ancolSymbole[ 6, 1]:=RGB(  0,128,128);
ancolSymbole[ 7, 1]:=RGB(192,192,192);
ancolSymbole[ 8, 1]:=RGB(  0,  0,  0);
ancolSymbole[ 9, 1]:=RGB(128,128,128); // Flag
ancolSymbole[10, 1]:=RGB(192,192,192); // ?
ancolSymbole[11, 1]:=RGB( 17, 19, 31); // Never selected.
ancolSymbole[12, 1]:=RGB(192,192,192); // to see
ancolSymbole[13, 1]:=RGB(192,192,192); // black bomb
ancolSymbole[14, 1]:=RGB(255,  0,  0); // bomb on red

// Second pixel color selection, many are useless
anXPosCol[2]:= 1;
anYPosCol[2]:= 1;

ancolSymbole[ 0, 2]:=RGB(192,192,192); // not changed
ancolSymbole[ 7, 2]:=RGB(192,192,192); // not changed
ancolSymbole[ 8, 2]:=RGB(192,192,192); // not changed
ancolSymbole[ 9, 2]:=RGB(255,255,255); // Flag
ancolSymbole[10, 2]:=RGB(255,255,255); // ?
ancolSymbole[12, 2]:=RGB(255,255,255); // to see
ancolSymbole[13, 2]:=RGB(192,192,192); // black bomb
ancolSymbole[14, 2]:=RGB(255,  0,  0); // bomb on red

// Third pixel color selection, many are useless
anXPosCol[3]:= 9;
anYPosCol[3]:= 5;

ancolSymbole[ 0, 3]:=RGB(192,192,192); // not changed
ancolSymbole[ 7, 3]:=RGB(192,192,192);
ancolSymbole[ 8, 3]:=RGB(192,192,192); // not changed
ancolSymbole[ 9, 3]:=RGB(128,  0,  0); // Flag
ancolSymbole[10, 3]:=RGB(  0,  0,255); // ?
ancolSymbole[12, 3]:=RGB(192,192,192); // to see
ancolSymbole[13, 3]:=RGB(128,128,128); // black bomb
ancolSymbole[14, 3]:=RGB(128,  0,  0); // bomb on red

// Last pixel color selection, many are useless
anXPosCol[4]:= 6;
anYPosCol[4]:= 4;
ancolSymbole[ 0, 3]:=RGB(192,192,192); // not changed
ancolSymbole[ 7, 4]:=RGB(128,  0,128); // 7

// Should be enough
end;  // TMineSolver.SetColorsWin95

procedure TMineSolver.SelectGameVersion;
//======================================
// Select the version of minesweeper.
// Actually, two version are programmed, Win95 and Win98.
var
  nRep        : integer;
  rRec        : RECT;
begin
nRep:=DialogExecMSG(30399, poWinMain);

if (nRep <> IDOK) then exit;

case nGameVersion of
   951 : SetColorsWin95;
   981 : SetColorsWin98;
   982 : begin // Use another methode to determine the content of the cells
         SetColorsWin98;
         nMethodContent:=0;
         end;
     0 : begin
         WriteWin(0, GetMsg(10481));
         end;
   end; // case

if (hWinSec > 0) and (nGameVersion > 0) then begin
   // Determine the number of cells
   GetClientRect(hWinSec, rRec);
   nXNumber:=(rRec.Right  - ptTopLeft.x - ptRightBottomMargin.x) div nXDelta;
   nYNumber:=(rRec.Bottom - ptTopLeft.y - ptRightBottomMargin.y) div nYDelta;
   end;
end;  // TMineSolver.SelectGameVersion

function TMineSolver.IsSecWindowDefined:boolean;
//==============================================
// if hWinSec <> 0  then return TRUE
// else display a dialog box explaining what to do, and return FALSE
var
  rRec        : RECT;
  nXNumberLoc,// to test if the size correspond to the number of cells
  nYNumberLoc : integer;

begin
IsSecWindowDefined:=FALSE; // default till tests have not passed

if (hWinSec = 0) then begin
   DialogExecMSG(20299, poWinMain);
   exit;
   end;

// Test that the dimensions of the window correspond to the number of cells.
GetClientRect(hWinSec, rRec);

// If incoherent values, put standard values
if (nXDelta <=0) then nXDelta:=16;
if (nYDelta <=0) then nYDelta:=16;

// Number of cells
nXNumberLoc:=(rRec.Right  - ptTopLeft.x - ptRightBottomMargin.x) div nXDelta;
nYNumberLoc:=(rRec.Bottom - ptTopLeft.y - ptRightBottomMargin.y) div nYDelta;

if (nXNumberLoc <> nXNumber) or (nYNumberLoc <> nYNumber) then begin
   // Sizes don't correspond, signal it.
   DialogExecMSG(20499, poWinMain);
   exit;
   end;

// The number of bombs could be tested, but the test can only
// tell if the number of bombs seems to be low or high.

IsSecWindowDefined:=TRUE;
end;  // TMineSolver.IsSecWindowDefined

procedure TMineSolver.MarkCell(nXP,nYP:integer; wColorP:word32);
//==============================================================
// Put a mark on the bottom right of the cell.
var
  hDCScreen   : HDC;
  ii,jj       : integer;
begin
// Get a device context of the screen
hDCScreen:=GetDC(0);

// Top left of the next cells, down right.
ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);
nXP:=ptTopLeftScreen.x + (nXP)* nXDelta;
nYP:=ptTopLeftScreen.y + (nYP)* nYDelta;

// Put a mark on pixels
for ii:=1 to 4 do for jj:=1 to 4 do
   SetPixel(hDCScreen, nXP-ii, nYP-jj, wColorP);

ReleaseDC(0, hDCScreen);
end;  // TMineSolver.MarkCell

procedure TMineSolver.GetCellPositions;
//=====================================
// Get the positions, sizes of all mine cells, and how many they are.
var
  nSecurity,
  nX,nY       : int32;
  wTime0      : word32;
  rRec        : RECT;
  hDCScreen   : HDC;
  colLast,    // Last pixel color
  colPixel    : TypeColor;
  nCountSame, // count the number of colors which where the same on a line
  nYBottomL,  // bottom of the last mine cell
  nXRightL    // right of the last mine cell
              : integer;
begin
if (hWinSec = 0) then begin
   DialogExecMSG(20299, poWinMain);
   exit;
   end;

// Initialisations
ptTopLeft.x:=0;
ptTopLeft.y :=0;
nXDelta :=0;
nYDelta :=0;
nXNumber :=0;
nYNumber :=0;

nXRightL:=0;
nYBottomL:=0;

// To get a starting point
GetWindowRect(hWinSec, rRec);
nX:=rRec.Right;
nY:=rRec.Bottom-20;

// Get a device context of the screen
hDCScreen:=GetDC(0);

// Get first pixel color
colLast:= GetPixel(hDCScreen, nX, nY);

nCountSame:=1;

nSecurity:=2000;
repeat
   nX:=nX - 1;

   // Moves the cursor
   SetCursorPos(nX,nY);

   colPixel:= GetPixel(hDCScreen, nX, nY);

   // count the number of pixels with the same colors
   if (colPixel = colLast) then inc(nCountSame)
   else begin
      // More than 10 pixels with the same colors means a cell.
      if (nCountSame >= 10) then begin
         // a mine cell has been passed through
         if     (nXRightL = 0) then nXRightL:=nX
         else if (nXDelta = 0) then nXDelta :=nXRightL - nX;

         inc(nXNumber);
         end;

      nCountSame:=1;
      colLast:=colPixel;
      end;

   // Wait to let us see the movement
   wTime0:=GetTickCount;
   repeat until wTime0 <> GetTickCount;

   until (nX <= rRec.Left) or (nSecurity <= 0);

// Loop to count the mine cells vertically
//"""""""""""""""""""""""""""""""""""""""""""
nX:=rRec.Left+20;
nY:=rRec.Bottom;

// Get first pixel color
colLast:= GetPixel(hDCScreen, nX, nY);

nCountSame:=1;

nSecurity:=2000;
repeat
   nY:=nY - 1;

   // Moves the cursor
   SetCursorPos(nX,nY);

   colPixel:= GetPixel(hDCScreen, nX, nY);

   if (colPixel = colLast) then inc(nCountSame)
   else begin
      if (nCountSame >= 10) then begin
         // a mine cell has been passed through
         if (nYBottomL = 0) then nYBottomL:=nY
         else if (nYDelta = 0) then nYDelta:=nYBottomL -(nY);

         inc(nYNumber);
         end;

      nCountSame:=1;
      colLast:=colPixel;
      end;

   // Wait to let us see the movement
   wTime0:=GetTickCount;
   repeat until wTime0 <> GetTickCount;

   until (nY <= rRec.Top+80) or (nSecurity <= 0);

ReleaseDC(0, hDCScreen);

// Position of the top left cell
ptTopLeft.x:=nXRightL  -1 - nXDelta * (nXNumber-1);
ptTopLeft.y:=nYBottomL -1 - nYDelta * (nYNumber-1);

// Moves the cursor to see the left top corner of the left top mine cell
SetCursorPos(ptTopLeft.x,ptTopLeft.y);

// Convert to have coordinate in client units.
ScreenToClient(hWinSec, ptTopLeft);

// Determine the margins
GetClientRect(hWinSec, rRec);
ptRightBottomMargin.x:= rRec.Right  - ptTopLeft.x - nXDelta*nXNumber;
ptRightBottomMargin.y:= rRec.Bottom - ptTopLeft.y - nYDelta*nYNumber;

//DBW (8, 'nXLeft, nXDelta, nXNumber=' + ITS5(ptTopLeft.x) + ITS5(nXDelta) + ITS5(nXNumber));
//DBW (8, 'nYTop, nYDelta, nYNumber =' + ITS5(ptTopLeft.y) + ITS5(nYDelta) + ITS5(nYNumber));
end;  // TMineSolver.GetCellPositions

procedure TMineSolver.GetSecWindowPosition;
//=========================================
// To get the position and size of the Second window, i.e. the window
// which was clicked before
var rRec : RECT;
begin
// Get the position and size of the Minesweeper window
GetWindowRect(hWinSec, rRec);
WriteWin(0, ITS5(rRec.Left) + ITS5(rRec.Top) + GetMsg(10501));
WriteWin(0, ITS5(rRec.Right - rRec.Left)
          + ITS5(rRec.Bottom - rRec.Top) + GetMsg(10502));
GetClientRect(hWinSec, rRec);
WriteWin(0, ITS5(rRec.Right) + ITS5(rRec.Bottom) + GetMsg(10503));

// If incoherent values, put standard values
if (nXDelta <=0) then nXDelta:=16;
if (nYDelta <=0) then nYDelta:=16;

// Number of cells
nXNumber:=(rRec.Right  - ptTopLeft.x - ptRightBottomMargin.x) div nXDelta;
nYNumber:=(rRec.Bottom - ptTopLeft.y - ptRightBottomMargin.y) div nYDelta;

// Assuming that we are in non cutom mode, sets the number of bombs.
nNbBombsTot:=0;
if (nXNumber =  8) and (nYNumber =  8) then nNbBombsTot:=10;
if (nXNumber =  9) and (nYNumber =  9) then nNbBombsTot:=10;
if (nXNumber = 16) and (nYNumber = 16) then nNbBombsTot:=40;
if (nXNumber = 30) and (nYNumber = 16) then nNbBombsTot:=99;

WriteWin(0, ITS5(nXNumber) + ITS5(nYNumber) + GetMsg(10504));
WriteWin(0, ITS(hWinSec) + GetMsg(10508));

if (nNbBombsTot > 0) then begin
   WriteWin(0, ITS5(nNbBombsTot) + GetMsg(10505));
   WriteWin(0, GetMsg(10506));
   end
else WriteWin(0, GetMsg(10507));

end;  // TMineSolver.GetSecWindowPosition

procedure TMineSolver.SimulateLeftClick;
//======================================
// Just to simulate a Left Click on the Minesweeper window.
var
  ptL         : TPoint;
begin
// Get the mouse position, to transmit it as a parameter to the message
GetCursorPos(@ptL);
ScreenToClient(hWinSec, ptL);

SendMessage(hWinSec, WM_LBUTTONDOWN, MK_LBUTTON, MakeLong(ptL.x, ptL.y));
SendMessage(hWinSec, WM_LBUTTONUP,            0, MakeLong(ptL.x, ptL.y));
end;  // TMineSolver.SimulateLeftClick

procedure TMineSolver.SimulateRightClick;
//=======================================
// Just to simulate a Left Click on the Minesweeper window.
var ptL         : TPOINT;
begin
// Get the mouse position, to transmit it as a parameter to the message
GetCursorPos(@ptL);
ScreenToClient(hWinSec, ptL);

SendMessage(hWinSec, WM_RBUTTONDOWN, MK_RBUTTON, MakeLong(ptL.x, ptL.y));
SendMessage(hWinSec, WM_RBUTTONUP,            0, MakeLong(ptL.x, ptL.y));
end;  // TMineSolver.SimulateRightClick

procedure TMineSolver.MoveMouseTo(nX2,nY2:integer);
//=================================================
// Moving the mouse cursor to the cell nX2, nY2,
// slowly of fast, depending on  nSpeed
var
  nSecurity,
  nXStep,nYStep,
  nX1,nY1,
  nX,nY,
  n2,n2Delta, // to adjust the mouse speed
  nn          : int32;
  wTime0      : word32;
  ptL         : TPOINT;
begin
// Change from the cell coordinate to the screen pixel cordinate
nX2:=4 + ptTopLeftScreen.x + (nX2-1)*nXDelta;
nY2:=4 + ptTopLeftScreen.y + (nY2-1)*nYDelta;

if (nSpeed >= 10) then begin
   // Solving fast, do not show the mouse cursor movements.
   SetCursorPos(nX2,nY2);
   exit;
   end;

GetCursorPos(@ptL);
nX1:=ptL.x;  nY1:=ptL.y;  // Starting position
nX:=nX1;     nY:=nY1;     // Starting position

if (nX1 = nX2) and (nY1 = nY2) then exit; // nothing to do

// To set the moving direction
if (nX2 > nX1) then nXStep:=1 else if (nX2 < nX1) then nXStep:=-1 else nXStep:=0;
if (nY2 > nY1) then nYStep:=1 else if (nY2 < nY1) then nYStep:=-1 else nYStep:=0;

n2:=100; // So we enter at least once in the waiting loop
nSecurity:=4000;
repeat
   // Move the mouse allong a strait line.
   // Must take care if (X1=X2) or if (Y1=Y2)
   if (abs((nY2-nY)*(nX2-nX1)) < abs((nX2-nX)*(nY2-nY1))) then nX:=nX + nXStep
   else if (abs((nY2-nY)*(nX2-nX1)) > abs((nX2-nX)*(nY2-nY1))) then nY:=nY + nYStep
   else if (nX <> nX2) then nX:=nX + nXStep
   else if (nY <> nY2) then nY:=nY + nYStep;
   // else do not move, the mouse cursor is already at destination.

   // Wait to let us see the movement
   nn:=10 - (5*nSpeed div 2) ;
   n2Delta:=100;
   if (nSpeed = 5) then n2Delta:=75;
   if (nSpeed = 6) then n2Delta:=50;
   if (nSpeed = 7) then n2Delta:=25;
   if (nSpeed = 8) then n2Delta:=10;
   if (nSpeed>= 9) then n2Delta:= 5;
   n2:=n2 + n2Delta;
   if (n2 >= 100) then begin
      n2:=n2-100;
      repeat
         // Wait nSpeed changes of TickCount.
         wTime0:=GetTickCount;
         repeat
            // Stop if the escape key has been pressed
            if (GetAsyncKeyState(VK_ESCAPE) < 0) or
               (GetAsyncKeyState(VK_SHIFT) < 0) then begin
               // Go to the destination quickly
               fQuitSolve:=TRUE;
               nn:=0;
               nX:=nX2;   nY:=nY2;
               end;

            // Accelerate the mouse speed or slow it down
            if (GetAsyncKeyState(VK_S) < 0) and fIsKeyUp then begin
               dec(nSpeed);
               if (nSpeed < 0) then nSpeed:=0;
               fIsKeyUp:=FALSE;
               end;
            if (GetAsyncKeyState(VK_F) < 0) and fIsKeyUp then begin
               inc(nSpeed);
               if (nSpeed > 10) then nSpeed:=10;
               fIsKeyUp:=FALSE;
               end;

            fIsKeyUP:= (GetAsyncKeyState(VK_F) >= 0) and
                       (GetAsyncKeyState(VK_S) >= 0);

            until wTime0 <> GetTickCount;
        dec(nn);
        until (nn <= 0); // or fQuitSolve
      end; // if

   // Moves the cursor
   SetCursorPos(nX,nY);

   dec(nSecurity);
   until ((nX = nX2) and (nY = nY2)) or (nSecurity <= 0);

// In over 1'000'000 tries, the problem never occured.
if (nSecurity <= 0) then begin
   ErrorWrite(1180, 'in MoveMouseTo');
   ErrorComment(1180, '***** PROBLEM, X1,Y1,X2,Y2=' +
                      ITS5(nX1) + ITS5(nY1) + ITS5(nX2) + ITS5(nY2));
   end;
end;  // TMineSolver.MoveMouseTo

procedure TMineSolver.ClickOnNewGame;
//===================================
// Click on the new game icone to begin a new minesweeper game
begin
// On the game I have, this reset the game.
SendMessage(hWinSec, WM_COMMAND, 510, 0);
end;  // TMineSolver.ClickOnNewGame

procedure TMineSolver.InitCellsData;
//==================================
// Initialisation to set that no cells have been visited, to see their content,
// and other array initialisation
var
  nInd, nX,nY       : Integer;
begin
// All cells must be visited
nInd:=0;
for nX:=1 to nXNumber do  for nY:=1 to nYNumber do begin
   inc(nInd);
   abXFree[nInd]:=nX;
   abYFree[nInd]:=nY;
   end;

nNbFreeCells:=nInd;

// Total number of bombs
nNbBombs:=nNbBombsTot;

// initialisation of the aanBO content.
// Just to simplify, have a border of ficive cells
for nX:=0 to nXNumber + 1 do begin
   aanBO[nX,0]          :=cteNoBomb; // Unexplored
   aanBO[nX,nYNumber+1] :=cteNoBomb;
   end;

for nY:=0 to nYNumber + 1 do begin
   aanBO[0,nY]          :=cteNoBomb; // Unexplored
   aanBO[nXNumber+1,nY] :=cteNoBomb;
   end;

for nX:=0 to nXNumber + 1 do
  for nY:=0 to nYNumber + 1 do begin
   aanTS[nX,nY]         := 8; // 8 unexplored cells around
   end;

// At the limits, the number of unexplored neighbour changes
for nX:=1 to nXNumber do begin
   aanTS[nX,         0]:= 3;
   aanTS[nX,         1]:= 5;
   aanTS[nX,nYNumber  ]:= 5;
   aanTS[nX,nYNumber+1]:= 3;
   end;

for nY:=1 to nYNumber do begin
   aanTS[         0,nY]:= 3;
   aanTS[         1,nY]:= 5;
   aanTS[nXNumber  ,nY]:= 5;
   aanTS[nXNumber+1,nY]:= 3;
   end;

aanTS[1,1]:= 3;  aanTS[nXNumber,1]:= 3;  aanTS[1,nYNumber]:= 3;
aanTS[nXNumber,nYNumber]:= 3;
aanTS[0,0]:= 1;                    aanTS[nXNumber+1,0]:= 1;
aanTS[nXNumber+1,nYNumber+1]:= 1;  aanTS[0,nYNumber+1]:= 1;
end;  // TMineSolver.InitCellsData

function TMineSolver.GetMineContentCol(hDCS:HDC; nXP,nYP:integer):integer;
//========================================================================
// Get the cell content, using colors
var
  colPixel1
              : TypeColor;
  nKept1,     // index of kept symboles
  nKept2,
  nLoop,nn
              : integer;
  anKeptL     : array[1..15] of integer;
  // List of selected symboles compatible with a the color found
begin
// Convert from cell coordinate to screen coordinate
// Top left of the cell
nXP:=ptTopLeftScreen.x + (nXP-1)* nXDelta;
nYP:=ptTopLeftScreen.y + (nYP-1)* nYDelta;

if fShowCellsOrgPixels then SetPixel(hDCS, nXP, nYP, RGB(0,255,192));

for nn:=1 to 15 do anKeptL[nn]:=nn-1;

nKept1:=15; // 14 possible symboles
nLoop:=0;
repeat
   inc(nLoop);
   // Get a first pixel to select some possible symboles
   colPixel1:= GetPixel(hDCS, anXPosCol[nLoop] + nXP, anYPosCol[nLoop] + nYP);

   // Select the symboles compatible with this color
   nKept2:=nKept1;
   nKept1:=0;
   for nn:=1 to nKept2 do begin
      if (colPixel1 = ancolSymbole[anKeptL[nn],nLoop]) then begin
         inc(nKept1);
         anKeptL[nKept1]:=anKeptL[nn];
         end;
      end;

   // If only one symbole selected, it is the content of the cell
   if (nKept1 = 1) then begin
      GetMineContentCol:=anSymbole[anKeptL[nKept1]];
      exit;
      end
   else if (nKept1 = 0) then begin
      // If no symboles selected, it is an error
      if (nWinGetMinCont1 >= 0) and (nWinGetMinCont1 <= 15) then begin
         WriteWin(nWinGetMinCont1, GetMsg(10551));
         WriteWin(nWinGetMinCont1, ITS5(nLoop) + GetMsg(10552));
         WriteWin(nWinGetMinCont1, ITS5(colPixel1.bRed) + ITS5(colPixel1.bGreen) +
                                   ITS5(colPixel1.bBlue) + GetMsg(10553));
         WriteWin(nWinGetMinCont1, ITS5(1 + (nXP - ptTopLeftScreen.x) div nXDelta) +
                                   ITS5(1 + (nYP - ptTopLeftScreen.y) div nYDelta) +
                                   GetMsg(10554));
         end;

      GetMineContentCol:=cteBoom; // Stop
      exit;
      end;

   until (nLoop >= 8);

// More than two symboles selected, its not normal.
if (nWinGetMinCont2 >= 0) and (nWinGetMinCont2 <= 15) then begin
   WriteWin(nWinGetMinCont2, GetMsg(10561));
   WriteWin(nWinGetMinCont2, GetMsg(10562));
   WriteWin(nWinGetMinCont2, ITS5(colPixel1.bRed) + ITS5(colPixel1.bGreen) +
                             ITS5(colPixel1.bBlue) + GetMsg(10563));
   WriteWin(nWinGetMinCont2, ITS5(1 + (nXP - ptTopLeftScreen.x) div nXDelta) +
                             ITS5(1 + (nYP - ptTopLeftScreen.y) div nYDelta) +
                             GetMsg(10564));
   end;

// Keep the first one, but an error is possible
GetMineContentCol:=anSymbole[anKeptL[1]];
end;  // TMineSolver.GetMineContentCol

function TMineSolver.GetMineContentWin98(hDCS:HDC; nXP,nYP:integer):integer;
//==========================================================================
// nXP, nYP = coordinate of a mine square.
// return the content of it.
// = cteToSee if it has not been explored, 0 if empty,
// = cteFlag  if a flag is in
// = cteBoom  if a bomb is in, and thus the game finished.
var
  colPixel1
 ,colPixel2
 ,colPixel3
 ,colPixel4
 ,colPixel5
            : TypeColor;

begin
// Convert from cell coordinate to screen coordinate
// Top left of the cell
nXP:=ptTopLeftScreen.x + (nXP-1)* nXDelta;
nYP:=ptTopLeftScreen.y + (nYP-1)* nYDelta;

// Get some pixels, to see the content of the mine square
colPixel1:= GetPixel(hDCS, 1+nXP, 1+nYP);
//SetPixel(hDCS, nXP, nYP, RGB(0,0,0));

if (colPixel1 <> colGray) then begin
   // It is either an unexplored square,
   // or a desactivated square
   // or a question mark
   // or the bombe which exploded
   colPixel2:= GetPixel(hDCS, 7+nXP, 5+nYP);

   if (colPixel2 = colGray) then begin
      colPixel3:= GetPixel(hDCS, 9+nXP, 5+nYP);

      if (colPixel3 = colGray) then GetMineContentWin98:=cteToSee // An unvisited cell
      else GetMineContentWin98:=cteQuestion;
      end
   else begin
      // A Flag or the bomb which exploded
      colPixel3:= GetPixel(hDCS, 6+nXP, 8+nYP);

      if (colPixel3 = colGray) then GetMineContentWin98:=cteFlag
      else GetMineContentWin98:=cteBoom; // The red bomb, which just exploded
      end;

   exit;
   end; // first pixel was not Gray

// Here we must determine the number of bombs around this square,
// i.e. read the number in the square, or return zero.

colPixel2:= GetPixel(hDCS,  6+nXP,  7+nYP);

if (colPixel2 = colGray) then begin
   // A 0,1,2 or 7
   colPixel3:= GetPixel(hDCS,  8+nXP,  7+nYP);

   if (colPixel3 = colGray) then begin
      // 0 or 7
      colPixel4:= GetPixel(hDCS,  9+nXP,  8+nYP);

      if (colPixel4 = colGray) then GetMineContentWin98:=0
      else                          GetMineContentWin98:=7;
      end
   else begin // 1 or 2
      colPixel4:= GetPixel(hDCS,  4+nXP,  4+nYP);

      if (colPixel4 = colGray) then GetMineContentWin98:=1
      else                          GetMineContentWin98:=2;
      end;
   end
else begin
   // A 3,4,5,6,8 or a bomb
   colPixel3:= GetPixel(hDCS,  3+nXP,  5+nYP);

   if (colPixel3 = colGray) then begin
      // 3,4 or a bomb
      colPixel4:= GetPixel(hDCS,  4+nXP,  8+nYP);

      if (colPixel4 = colGray) then GetMineContentWin98:=3
      else begin
         colPixel5:= GetPixel(hDCS,  7+nXP,  10+nYP);

         if (colPixel5 = colGray) then GetMineContentWin98:=4
         else                          GetMineContentWin98:=cteBoom;
         end
      end
   else begin
      // 5,6,8
      colPixel4:= GetPixel(hDCS,  11+nXP,  6+nYP);

      if (colPixel4 = colGray) then begin
         // 5 or 6
         colPixel5:= GetPixel(hDCS,  3+nXP,  10+nYP);

         if (colPixel5 = colGray) then GetMineContentWin98:=5
         else                          GetMineContentWin98:=6;
         end
      else GetMineContentWin98:=8
      end;
   end;
end;  // TMineSolver.GetMineContentWin98

procedure TMineSolver.GetMineSweeperContent(wFlagsP:word32);
//==========================================================
// Get the content of the complete minesweeper window
// wFlagsP =1 => show the content in the debug window.
// wFlagsP = 6 => Cells with Flags and Question are set to be not visited
var
  nContent,   // Content of a mine square
  nX,nY,
  ii,jj       : integer;
  hDCScreen   : HDC;
  strS        : ShortString;
begin
if (nXDelta = 0) then GetCellPositions;

for nX:=0 to nXNumber + 1 do
  for nY:=0 to nYNumber + 1 do begin
   aanBO[nX,nY] := 0;
   aanTS[nX,nY] := 0; // Unexplored
   end;

// initialisation of the aanBO content.
for nX:=0 to nXNumber + 1 do begin
   aanBO[nX,0]          :=cteNoBomb; // Unexplored
   aanBO[nX,nYNumber+1] :=cteNoBomb;
   end;

for nY:=0 to nYNumber + 1 do begin
   aanBO[0,nY]          :=cteNoBomb; // Unexplored
   aanBO[nXNumber+1,nY] :=cteNoBomb;
   end;

{for nX:=1 to nXNumber do begin
   aanBO[nX,1]        :=0; // Unexplored
   aanBO[nX,nYNumber] :=0;
   end;
}
// Get a device context of the screen
hDCScreen:=GetDC(0);

ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

// Reference Gray color
colGray:=GetPixel(hDCScreen, 2+ptTopLeftScreen.x, 2+ptTopLeftScreen.y);

// Loop on all mine cells to see their content
nNbBombs:=nNbBombsTot;
nNbFreeCells:=0;
nY:=1;
repeat
   nX:=1;

   // Loop on a row
   repeat
      // Get the content of the cell (nX, nY)
      // Two different methodes can be used to get the content of the game,
      // depending on the game version.
      if (nMethodContent = 1) then nContent:=GetMineContentCol(hDCScreen, nX, nY)
      else                         nContent:=GetMineContentWin98(hDCScreen, nX, nY);

      // Flagged cells are set to be unvisited.
      if (wFlagsP = 6) and
         ((nContent = cteFlag) or (nContent = cteQuestion)) then
         nContent := cteToSee;

      // To count the neighbour cell to be still explored
      if (nContent = cteToSee) then begin
         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aanTS[ii,jj]:=aanTS[ii,jj] + 1;

         aanTS[nX,nY]:=aanTS[nX,nY] - 1;

         // One more unvisited cell
         inc(nNbFreeCells);
         abXFree[nNbFreeCells]:=nX;
         abYFree[nNbFreeCells]:=nY;
         end
      else if (nContent = cteFlag) then begin
         // To count the neighbour cell having a marked bomb
         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aanBO[ii,jj]:=aanBO[ii,jj] - 1;

         // One less unvisited cell having a bomb
         dec(nNbBombs);
         end
      else if (nContent = cteBoom) then inc(nNbBombExplosed); // Lost

      // Content of this cell
      aanBO[nX,nY]:=aanBO[nX,nY] + nContent;

      inc(nX);
      until (nX > nXNumber);

   inc(nY);
   until (nY > nYNumber);

ReleaseDC(0, hDCScreen);

if (wFlagsP <> 1) then exit; // no info in the display window

if (nWinSeeMineContent < 0) or (nWinSeeMineContent > 15) then exit;

WriteWin(nWinSeeMineContent, '--------------------------------------------');
// To show what has been found in the cells
for nY:=1 to nYNumber do begin
   strS:='|';
   for nX:=1 to nXNumber do begin
      if (aanBO[nX,nY] >= cteBoom-9) then // A bomb
        strS:=strS + '*'
      else if (aanBO[nX,nY] = cteToSee) then // An untouched cell
        strS:=strS + '.'
      else if (aanBO[nX,nY] = cteToSee-1) then // An untouched cell
        strS:=strS + ','
      else if (aanBO[nX,nY] >= cteToSee-2) then // An untouched cell
        strS:=strS + ':'
      else if (aanBO[nX,nY] >= cteToSee-9) then // An untouched cell
        strS:=strS + ';'
      else if (aanBO[nX,nY] >= cteQuestion-9) then // A ?
        strS:=strS + '?'
      else if (aanBO[nX,nY] >= cteFlag-9) then // A flgs is in
        strS:=strS + 'F'
      else
        strS:=strS + IntToStr(aanBO[nX,nY],1);

      strS:=strS + IntToStr(aanTS[nX,nY],1) + ' ';

      end; // for nX

   WriteWin(nWinSeeMineContent, strS + '|');
   end; // for nY
WriteWin(nWinSeeMineContent, ITS5(nNbBombs) + ITS5(nNbFreeCells) + GetMsg(10601));
end;  // TMineSolver.GetMineSweeperContent

procedure TMineSolver.UpdateMineSweeperContent;
//=============================================
// The content of the mineSweeper is supposed to have been set right
// before the last click, from it, update the information about
// the content.
var
  nContent,   // Content of a mine cell
  nn,
  nX,nY,
  ii,jj       : integer;
  hDCScreen   : HDC;
//  strS        : ShortString;
begin
// Get a device context of the screen
hDCScreen:=GetDC(0);

ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

// Reference Gray color
colGray:=GetPixel(hDCScreen, 2+ptTopLeftScreen.x, 2+ptTopLeftScreen.y);

// Loop on all unvisited cells to see their content
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

      // Get the content of the cell (nX, nY)
      // Two different methodes can be used to get the content of the game,
      // depending on the game version.
      if (nMethodContent = 1) then nContent:=GetMineContentCol(hDCScreen, nX, nY)
      else                         nContent:=GetMineContentWin98(hDCScreen, nX, nY);

      // To count the neighbour cell to be still explored
      if (nContent <> cteToSee) then begin
         // Content of this cell
         aanBO[nX,nY]:=aanBO[nX,nY] + nContent - cteToSee;

         if (nContent = cteBoom) then inc(nNbBombExplosed); // Lost

         // Unvisited (=cteToSee) neighbours
         aanTS[nX,nY]:=aanTS[nX,nY] + 1;

         for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do
           aanTS[ii,jj]:=aanTS[ii,jj] - 1;

         // Copy the last cell of the array to the one which has been visited
         abXFree[nn]:=abXFree[nNbFreeCells];
         abYFree[nn]:=abYFree[nNbFreeCells];

         // One less unvisited cell
         dec(nNbFreeCells);

         dec(nn); // this cell is still unvisited
         end;

   until (nn >= nNbFreeCells);

ReleaseDC(0, hDCScreen);
end;  // TMineSolver.UpdateMineSweeperContent

function TMineSolver.PutFlags(wFlagsP:word32):integer;
//====================================================
// Put Flags on cells where it is sure that a bomb existe.
// Return the number of flag added.
// wFlagsP = 0 => normal playing.
// wFlagsP = 1 => do not play, only count the number of cells which could
// wFlagsP = 2 => do not play, only mark the cells which have a bomb
// are detected having a bombs and not being flagged.
var
  nn,
  nCount,
  nX,nY,
  ii,jj
              : integer;
begin
if (nNbFreeCells <= 0) then begin
   PutFlags:=0;
   exit;
   end;

ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

// Loop on all unexplored cells, to see if they contain a bomb and
// thus a flag must be set.
nCount:=0;
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

   if (aanTS[nX,nY] < 8) then begin
      // Loop on all neighbour to see if the cell has a bomb
      ii:=nX-2;
      repeat
         inc(ii);
         jj:=nY-2;
         repeat
            inc(jj);
            if (ii=nX) and (jj=nY) then inc(jj); // Skip the (nX,nY) cell

            if (aanBO[ii,jj] = aanTS[ii,jj]) then begin
               // the (nX,nY) cell has a bomb, mark it
               if (wFlagsP = 0) then begin
                  MoveMouseTo(nX, nY);
                  SimulateRightClick;
                  end
               else if (wFlagsP = 2) then begin
                  MarkCell(nX, nY, RGB(255,0,0));
                  end;

               // Mark that neighbour has one more neighbour with a bomb
               // and one less unvisited neighbour
               aanTS[nX,nY]:=aanTS[nX,nY] + 1;
               for ii:=nX-1 to nX+1 do  for jj:=nY-1 to nY+1 do begin
                  aanBO[ii,jj]:=aanBO[ii,jj] - 1;
                  aanTS[ii,jj]:=aanTS[ii,jj] - 1;
                  end; // for

               aanBO[nX,nY]:=cteFlag;
               dec(nNbBombs);

               // Copy the last cell of the array to the one which has been visited
               abXFree[nn]:=abXFree[nNbFreeCells];
               abYFree[nn]:=abYFree[nNbFreeCells];

               // One less unvisited cell
               dec(nNbFreeCells);

               dec(nn); // this cell is still unvisited

               inc(nCount);

               // no more searching necessary, quit the repeat loops
               ii:=nX+1;
               jj:=nY+1;
               end; // if
            until (jj > nY);
         until (ii > nX);
      end; // if (aanTS[nX,nY] < 8) then begin
   until (nn >= nNbFreeCells) or fQuitSolve;

// Nothing to update, it is already up to date.
//UpdateMineSweeperContent;

PutFlags:=nCount;
end;  // TMineSolver.PutFlags

procedure TMineSolver.StartingExplore;
//====================================
// When we beging, start with some special cells.
var
  nX,nY,      // where to click
  nDX,nDY     // to go on cells neighbour of the previous
              : integer;
  fQuitRepeat : boolean;
begin
if (aanBO[1,1] <> cteToSee) then  exit;  // not an empty field

ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

// First cell to explore  (1,1)
nX:=1;   nY:=1;
nDX:=2;  nDY:=2;

// Try to find a good starting corner
fQuitRepeat:=FALSE;
repeat
   MoveMouseTo(nX, nY);
   SimulateLeftClick;
   UpdateMineSweeperContent;

   if (aanBO[nX,nY] = 0) then exit; // good start.
   if (nNbBombExplosed > 0) then exit;

   if (aanBO[nX,nY] = 3) then begin
      // Click near the previous cell  (1,3)
      nY:=nY + nDY;
      MoveMouseTo(nX, nY);
      SimulateLeftClick;
      UpdateMineSweeperContent;
      if (aanBO[nX,nY] <= 2) then exit; // good second try
      if (nNbBombExplosed > 0) then exit;

      if (aanBO[nX,nY] = 5) then begin
         // Click again near the previous cell  (1,5)
         nY:=nY + nDY;
         MoveMouseTo(nX, nY);
         SimulateLeftClick;
         UpdateMineSweeperContent;
         if (aanBO[nX,nY] <= 2) then exit; // good second try
         if (nNbBombExplosed > 0) then exit;

         if (aanBO[nX,nY] = 5) then begin
            // Click again near the previous cell  (1,7)
            nY:=nY + nDY;
            MoveMouseTo(nX, nY);
            SimulateLeftClick;
            UpdateMineSweeperContent;
            if (aanBO[nX,nY] <= 2) then exit; // good second try
            if (nNbBombExplosed > 0) then exit;
            end; // (1,7)
         end; // (1,5)

      // Click near the previous cell  (3,1)
      nX:=nX + nDX;
      nY:=nY - nDY;
      MoveMouseTo(nX, nY);
      SimulateLeftClick;
      UpdateMineSweeperContent;
      if (aanBO[nX,nY] <= 2) then exit; // good second try
      if (nNbBombExplosed > 0) then exit;

      if (aanBO[nX,nY] = 5) then begin
         // Click again near the previous cell  (5,1)
         nX:=nX + nDX;
         MoveMouseTo(nX, nY);
         SimulateLeftClick;
         UpdateMineSweeperContent;
         if (aanBO[nX,nY] <= 2) then exit; // good second try
         if (nNbBombExplosed > 0) then exit;

         if (aanBO[nX,nY] = 5) then begin
            // Click again near the previous cell  (7,1)
            nX:=nX + nDX;
            MoveMouseTo(nX, nY);
            SimulateLeftClick;
            UpdateMineSweeperContent;
            if (aanBO[nX,nY] <= 2) then exit; // good second try
            if (nNbBombExplosed > 0) then exit;
            end; // (7,1)
         end; // (5,1)
      end; // (3,1)

   // Try in another corner
   if (nY = 1) then begin
      nY:=nYNumber;
      nDY:=-2;
      end
   else if (nX = 1) then begin
      nX:=nXNumber;
      nDX:=-2;
      nY:=1;
      nDY:= 2;
      end
   else begin
      nX:=nX + nDX;
      MoveMouseTo(nX, nY);
      SimulateLeftClick;
      UpdateMineSweeperContent;

      fQuitRepeat:=TRUE;
      end;

   until fQuitRepeat;

// Here other tries could be tempted.

end;  // TMineSolver.StartingExplore

function TMineSolver.SaftyExplore(wFlagsP:word32):integer;
//========================================================
// Click on cells which have no bombs for sure
// Return the number of cells which have been clicked.
// wFlagsP = 0 => normal playing.
// wFlagsP = 1 => do not play, only count the number of cells which have no bombs
// wFlagsP = 2 => do not play, only mark the cells which have no bombs
var
  nCount,     // Count how many cells have been clicked
  nn,
  nX,nY,
  ii,jj       : integer;
begin
ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

nCount:=0;

// Loop on all unexplored cells, to see if they contain a bomb and
// thus a flag must be set.
nn:=0;
repeat
   inc(nn);
   nX:=abXFree[nn];
   nY:=abYFree[nn];

   if (aanTS[nX,nY] < 8) then begin
      // Loop on all neighbour to see if the cell has no bomb for sure
      ii:=nX-2;
      repeat
         inc(ii);
         jj:=nY-2;
         repeat
            inc(jj);
            if (ii=nX) and (jj=nY) then inc(jj); // Skip the (nX,nY) cell

            if (aanBO[ii,jj] = 0) then begin
               // neighbour of this cell have no bombs for sure
               // Particulary the cell (nX, nY) has no bomb
               if (wFlagsP = 0) then begin
                  MoveMouseTo(nX, nY);
                  SimulateLeftClick;
                  end
               else if (wFlagsP = 2) then begin
                  MarkCell(nX, nY, RGB(0,0,0));
                  end;

               inc(nCount);

               // no more searching necessary, quit the repeat loops
               ii:=nX+1;
               jj:=nY+1;
               end; // if
            until (jj > nY);
         until (ii > nX);
      end; // if (aanTS[nX,nY] < 8) then begin
   until (nn >= nNbFreeCells) or fQuitSolve;

// The minesweeper window changed, update the information we have about it.
if (nCount > 0) and (wFlagsP = 0) then UpdateMineSweeperContent;
SaftyExplore:=nCount;
end;  // TMineSolver.SaftyExplore

procedure TMineSolver.SaveValue(pnValue:Pinteger);
//================================================
// Save in  apnModif[nIndModif]  and anModifValue[nIndModif]
// the pointer  pnValue,  and the content of pnValue^
//const wIndMax : word32 = 0;
begin
{// For debugging ----------------------------------
if (wIndMax < word32(nIndModif)) then begin
    wIndMax:=nIndModif;
    WriteWin(2, 'wIndMax=' + ITS5(wIndMax));
    end;
}//--------------------------------------------

inc(nIndModif);
apnModif[nIndModif]:=pnValue;
anModifValue[nIndModif]:=pnValue^;
end;  // TMineSolver.SaveValue

function TMineSolver.SetFlagTo(nXP,nYP:integer):integer;
//======================================================
label BeginWithNoBomb;
var
  nX,nY,
  nX0,nY0,
  ii,jj,
  nn
              : integer;
  anXmem,anYmem // to memorise which cells have been modified
              : array[1..100] of integer;
  nInd0,
  nInd1,
  nInd2       : integer; // number of modified cells

begin
SetFlagTo:=100; // Default return value, indicating no contradiction found

// First cell, being marked having a bomb
anXmem[1]:=abs(nXP);    anYmem[1]:=nYP;
nInd1:=0;
nInd2:=1;

if (nIndModif >= 0) then begin
   SaveValue(@aanBObis[abs(nXP),nYP]);
   dec(nNbFreeCellsBis);
   if (nXP >= 0) then dec(nNbBombsBis);
   end;

if (nXP < 0) then begin
   // Code to say to begin with a cell marked with no bomb
   nXP:=-nXP;
   aanBObis[nXP,nYP]:=cteNoBomb;
   goto BeginWithNoBomb;
   end
else aanBObis[nXP,nYP]:=cteFlag;

repeat
   // Treat all cells being marked having a bomb
   //""""""""""""""""""""""""""""""""""""""""""""
   nInd0:=nInd1;
   nInd1:=nInd2;
   for nn:=nInd0+1 to nInd1 do begin
      nX0:=anXmem[nn];
      nY0:=anYmem[nn];

      for nX:=nX0-1 to nX0+1 do  for nY:=nY0-1 to nY0+1 do
       if (nX <> nX0) or (nY <> nY0) then begin // Skip the cell (nX0,nY0)
         if (aanBObis[nX,nY] = 0) then begin
            // Contradiction found, the cell (nXP, nYP) cannot have a bomb
            SetFlagTo:=-1;
            exit;
            end;

         // Neighbour has one less cell with a bomb
         // and one less unvisited cell
         if (nIndModif >= 0) then begin
            SaveValue(@aanBObis[nX,nY]);
            SaveValue(@aanTSbis[nX,nY]);
            end;
         aanBObis[nX,nY]:=aanBObis[nX,nY] - 1;
         aanTSbis[nX,nY]:=aanTSbis[nX,nY] - 1;

         if (aanBObis[nX,nY] = 0) then begin
            // All unvisited neighbour of (nX,nY) have no bombs
            // Memorise them to treat them later
            for ii:=nX-1 to nX+1 do for jj:=nY-1 to nY+1 do
             if ((ii <> nX) or (jj <> nY)) and
                (aanBObis[ii,jj] >= cteToSee-9) then begin
               if (nInd2 > 100) then exit; // search too long
               inc(nInd2);
               anXmem[nInd2]:=ii; // Marked as having NO bomb
               anYmem[nInd2]:=jj;

               if (nIndModif >= 0) then begin
                  SaveValue(@aanBObis[ii,jj]);
                  dec(nNbFreeCellsBis);
                  end;
               aanBObis[ii,jj]:=cteNoBomb;
               end; // for .. if
            end; // if (aanBObis[nX,nY] = 0) then begin
         end; // for nX:=...
      end; // for nn:=...

BeginWithNoBomb:;
   // Treat all cells having been marked without bombs
   //"""""""""""""""""""""""""""""""""""""""""""""""""
   nInd0:=nInd1;
   nInd1:=nInd2;
   for nn:=nInd0+1 to nInd1 do begin
      nX0:=anXmem[nn];
      nY0:=anYmem[nn];

      for nX:=nX0-1 to nX0+1 do  for nY:=nY0-1 to nY0+1 do
       if (nX <> nX0) or (nY <> nY0) then begin
         if (aanBObis[nX,nY] = aanTSbis[nX,nY]) then begin
            // Contradiction...
            SetFlagTo:=-1;
            exit;
            end;

         if (nIndModif >= 0) then begin
            SaveValue(@aanTSbis[nX,nY]);
            end;
         aanTSbis[nX,nY]:=aanTSbis[nX,nY] -1;

         if (aanBObis[nX,nY] = aanTSbis[nX,nY]) then begin
            // all unvisited neighbour cell have a bomb.
            // Memorise them to treat them later
            for ii:=nX-1 to nX+1 do for jj:=nY-1 to nY+1 do
             if ((ii <> nX) or (jj <> nY)) and
                (aanBObis[ii,jj] >= cteToSee-9) then begin
               if (nInd2 > 100) then exit; // search too long
               inc(nInd2);
               anXmem[nInd2]:=ii; // Marked as having a bomb
               anYmem[nInd2]:=jj;

               if (nIndModif >= 0) then begin
                  SaveValue(@aanBObis[ii,jj]);
                  dec(nNbFreeCellsBis);
                  if (nXP >= 0) then dec(nNbBombsBis);
                  end;
               aanBObis[ii,jj]:=cteFlag;
               end;  // for .. if
            end; // if (aanBObis[nX,nY] = aanTSbis[nX,nY])
         end; // for nX:=...
      end; // for nn:=...

   until (nInd2 = nInd1);

SetFlagTo:=nInd2;
end;  // TMineSolver.SetFlagTo

function TMineSolver.PutFlagsMore(wFlagsP:word32):integer;
//========================================================
// Put Flags on cells where it is sure that a bomb existe.
// Use more deduction than in PutFlags to find these cells
// Return the number of flag added.
// wFlagsP = 0 => normal playing.
// wFlagsP = 1 => do not play, only count the number of cells which could
// wFlagsP = 2 => do not play, only mark the cells which have a bomb
// are detected having a bombs and not being flagged.
var
  nn,         // To get at random an unvisited cell
  nCount,
  ii,jj,
  nXL,nYL     : integer;
begin
nCount:=0; // If no more bombs has been found

// Do not save values
nIndModif:=-1;

// If nothing found, Try to find cells having a bomb for sure.
nn:=0;
repeat
   inc(nn);
   nXL:=abXFree[nn];
   nYL:=abYFree[nn];

   if (aanTS[nXL,nYL] <> 8) then begin
      aanBObis:=aanBO;
      aanTSbis:=aanTS;
      // Try to see if there is a contradiction setting a cteNoBomb there
      if (SetFlagTo(-nXL,nYL) = -1) then begin
         // A cell with a bomb has been found
         if (wFlagsP = 0) then begin
            // Mark this cell
            MoveMouseTo(nXL, nYL);
            SimulateRightClick;
            end
         else if (wFlagsP = 2) then begin
            MarkCell(nXL, nYL, RGB(255,0,0));
            end;

         // Mark that neighbour has one more neighbour with a bomb
         // and one less unvisited neighbour
         aanTS[nXL,nYL]:=aanTS[nXL,nYL] + 1;
         for ii:=nXL-1 to nXL+1 do  for jj:=nYL-1 to nYL+1 do begin
             aanBO[ii,jj]:=aanBO[ii,jj] - 1;
             aanTS[ii,jj]:=aanTS[ii,jj] - 1;
             end; // for

         aanBO[nXL,nYL]:=cteFlag;
         dec(nNbBombs);

         // Copy the last cell of the array to the one which has been visited
         abXFree[nn]:=abXFree[nNbFreeCells];
         abYFree[nn]:=abYFree[nNbFreeCells];

         // One less unvisited cell
         dec(nNbFreeCells);

         dec(nn); // this cell is still unvisited

         inc(nCount); // Indicate that a bomb has been found
         end;
      end;
   until (nn >= nNbFreeCells) or fQuitSolve;

PutFlagsMore:=nCount;
end;  // TMineSolver.PutFlagsMore

function TMineSolver.SaftyExploreMore(wFlagsP:word32):integer;
//============================================================
// Click on cells which have no bombs for sure
// But the searching of these cells is more sophisticated than SaftyExplore
// wFlagsP = 0 => normal playing.
// wFlagsP = 1 => do not play, only count the number of cells which could
// wFlagsP = 2 => do not play, only mark the cells which have no bombs
// are detected having no bombs.
var
  nn,         // To get at random an unvisited cell
  nCount,
  nFlagged,   // number of cells found to have a bomb
  nXL,nYL     : integer;
  fTryFlags   : boolean;
begin
ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

// Do not save values
nIndModif:=-1;

// Try first to find cells without bombs, if no success, then try
// to find cells with bombs.
nFlagged:=0;
fTryFlags:=TRUE;
repeat
 // Try to find cells having NO bomb for sure.
 nCount:=0; // If no more bombs has been found
 nn:=0;
 repeat
   inc(nn);
   nXL:=abXFree[nn];
   nYL:=abYFree[nn];

   if (aanTS[nXL,nYL] <> 8) then begin
      aanBObis:=aanBO;
      aanTSbis:=aanTS;
      // Try to see if there is a contradiction setting a flag there
      if (SetFlagTo(nXL,nYL) = -1) then begin
         // Contradiction found, thus there are no bombs on this cell
         if (wFlagsP = 0) then begin
            // Click on this cell
            MoveMouseTo(nXL, nYL);
            SimulateLeftClick;
            end
         else if (wFlagsP = 2) then begin
            MarkCell(nXL, nYL, RGB(0,0,0));
            end;

         inc(nCount); // Indicate that a bomb has been found
         end;
      end;
   until (nn >= nNbFreeCells);

 if (nCount = 0) and fTryFlags then begin
    // Try to find cells with bombs for sure
    nFlagged:=PutFlagsMore(wFlagsP);
    fTryFlags:=FALSE;
    end;
 until (nCount > 0) or (nFlagged = 0) or (not fTryFlags);

// The minesweeper window changed, update the information we have about it.
if (nCount > 0) and (wFlagsP = 0) then UpdateMineSweeperContent;
SaftyExploreMore:=nCount;
end;  // TMineSolver.SaftyExploreMore

procedure TMineSolver.SortFreeCells{(nFirstP,nLastP,nPivotP:integer)};
//==================================================================
// Sort the abXFree,abYFree arrays such that cells having few unvisited
// neighbour come first.
var
  nTemp,
  nL,nR       : integer;
  fFoundL     : boolean;

begin
// First sorting, separat cells having a neighbour containing information
// from cells having no neighbour containing information
nL:=1;
nR:=nNbFreeCells;

repeat
   dec(nL);
   repeat
      fFoundL:=FALSE;
      inc(nL);

      // Search a cell having no neighbour containing information, i.e.
      // having only unvisited neighbours
      if (nL > nNbFreeCells) then fFoundL:=TRUE
      else if (abXFree[nL] > 1) and (abXFree[nL] < nXNumber) then begin
         // Not the first neither the last line
         if (abYFree[nL] > 1) and (abYFree[nL] < nYNumber) then begin
            // Not the first neither the last column
            if (aanTS[abXFree[nL],abYFree[nL]] =
                8 + aanBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end
         else begin // In the first or last column, but not on first or last
            if (aanTS[abXFree[nL],abYFree[nL]] =
                5 + aanBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end;
         end
      else begin
         // In the first or last line
         if (abYFree[nL] > 1) and (abYFree[nL] < nYNumber) then begin
            // Not the first neither the last column
            if (aanTS[abXFree[nL],abYFree[nL]] =
                5 + aanBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end
         else begin // In the corners
            if (aanTS[abXFree[nL],abYFree[nL]] =
                3 + aanBO[abXFree[nL],abYFree[nL]] - cteToSee) then fFoundL:=TRUE;
            end;
         end;
      until fFoundL;

   inc(nR);
   repeat
      fFoundL:=TRUE;
      dec(nR);

      // Search a cell having a neighbour containing information
      if (nR < 1) then fFoundL:=TRUE
      else if (abXFree[nR] > 1) and (abXFree[nR] < nXNumber) then begin
         // Not the first neither the last line
         if (abYFree[nR] > 1) and (abYFree[nR] < nYNumber) then begin
            // Not the first neither the last column
            if (aanTS[abXFree[nR],abYFree[nR]] =
                8 + aanBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end
         else begin // In the first or last column
            if (aanTS[abXFree[nR],abYFree[nR]] =
                5 + aanBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end;
         end
      else begin
         // In the first or last line
         if (abYFree[nR] > 1) and (abYFree[nR] < nYNumber) then begin
            // Not the first neither the last column
            if (aanTS[abXFree[nR],abYFree[nR]] =
                5 + aanBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end
         else begin // In the corners
            if (aanTS[abXFree[nR],abYFree[nR]] =
                3 + aanBO[abXFree[nR],abYFree[nR]] - cteToSee) then fFoundL:=FALSE;
            end;
         end;
      until fFoundL;

   if (nL < nR) then begin
      // Exange the index, to have first, cells having neighbour containing information
      nTemp       :=abXFree[nR];
      abXFree[nR] :=abXFree[nL];
      abXFree[nL] :=nTemp;
      nTemp       :=abYFree[nR];
      abYFree[nR] :=abYFree[nL];
      abYFree[nL] :=nTemp;
      inc(nL);
      dec(nR);
      end;

   until (nL > nR);

// Number of interresting cells. No deduction can be made on the others
// without taking the number of bombs into account.
nNbFreeCellsInfo:=nR;

//DBW (1, 'nbFree=' + ITS5(nNbFreeCellsInfo) + ITS5(nNbFreeCells));

// Indexes from  1  to  nR  are from cells having neighbour containing information
{for nL:=1 to nR do begin
   DBW (1, 'info:' + ITS5(abXFree[nL]) + ITS5(abYFree[nL])
                  + ITS5(aanTS[abXFree[nL],abYFree[nL]]));
   end;
}
end;  // TMineSolver.SortFreeCells
{
procedure TMineSolver.ConfigStore(nNbConfigWI:integer);
//=====================================================
// Store in  arConfig  the configuration compatible with the actual state
// of the game.
// nNbConfigWI = number of configurations of the cells Without Info
var
  ii,jj,
  nNbBombL, // number of neighbour cells with bombs
  nn : integer;

begin
nNbConfigWI:=nNbConfigWI; // to avoid a hint
inc(nNbConfig);

for nn:=1 to nNbFreeCells do with arConfig[nNbConfig] do begin
   // Mark the cell having a bomb
   if (aanBOBis[abXFree[nn],abYFree[nn]] <= cteFlag) then anC[nn]:=09
   else begin
      // Count the number of neighbour having a bomb
      nNbBombL:=0;
      for ii:=abXFree[nn]-1 to abXFree[nn]+1 do
       for jj:=abYFree[nn]-1 to abYFree[nn]+1 do begin
         if (aanBOBis[ii,jj] <= cteFlag) and
            (aanBOBis[ii,jj] > 10) then inc(nNbBombL);
         end;

      anC[nn]:=nNbBombL;
      end;

//DBW(1024, ITS5(abXFree[nn]) + ITS5(abYFree[nn]) + ITS5(anC[nStore]) +
//       ITS5(aanBOBis[abXFree[nn], abYFree[nn]])  );

   end; // for nn...
end;  // TMineSolver.ConfigStore

function TMineSolver.NbLostConfig(
//================================
       nCell_0:integer; // cell on which we click
   var anConfList:array of integer; // list of possible configurations
       nNbConfList:integer; // number of possible configurations
   var anCellFree:array of integer; // list of unvisited cells of the configurations
       nNbCellFree:integer; // number of unvisited cells
       nNbLostConfigMax:integer // maximum loosing configurations accepted.
   ):integer;
// return the number of loosing configurations if the cell "nCell_0"
// is clicked on.
// Use NbWinConfig recursively to compute its value.
// Can be very time consuming
// We know a cell where clicking on it will give at most
// nNbLostGameMax loosing config. Thus it is useless to test cells with more
// than this number of loosing configurations
var
  aanPart     : array[0..8,0..MaxConfigStore] of integer;
  // Partition list of compatible config.
  // A lot of memory is wasted... but its enough complicate to not bother about it
  anIndPart   : array[0..09] of integer;
              // Number of element in each partitions.
  anCellFreeNew : array[0..MaxFreeCellStore] of integer;
              // List of unvisited cells.
  nNbLostGame, // number of loosing config.
  nNbLostGameAdd, // number of loosing games to add to the loosing config.
  nConf,      // To loop among the configurations
  nCell,      // To loop among the cells
  nCL,        // Content of the cell, for each configurations
  nInd,       // Temporary, index of the partitions.
  nLost,      // Count the number of loosing configurations for on partition and one cell
  nCellBest,  // Cell which has the lowest probability to loose imediatly
  nLostBest,  // Number of config. having a bomb in the cell "nCell"
  nn,n2
              : integer;
begin
// First make a partition of the compatibles configurations with
// the content of the cell  nCell_0
for nCL:=0 to 09 do anIndPart[nCL]:=0;

for nConf:=1 to nNbConfList do begin
   // Number of neighbour cell having a bomb
   nCL:=arConfig[anConfList[nConf]].anC[nCell_0];

   // One more configuration in the partition  nCL
   inc(anIndPart[nCL]);
   nInd:=anIndPart[nCL];
   if (nCL <= 8) then aanPart[nCL,nInd]:=anConfList[nConf];
   end;

// anIndPart[09] = number of config having a bombs in the cell  nCell_0
nNbLostGame:=anIndPart[09];

if (nNbLostGame >= nNbLostConfigMax) then begin
   // bad cell to click on.
   NbLostConfig:=nNbLostGame;
   exit;
   end;

// List of unvisited cells
n2:=0;
for nn:=1 to nNbCellFree do begin
   if (anCellFree[nn] <> nCell_0) then begin
      inc(n2);
      anCellFreeNew[n2]:=anCellFree[nn];
      end;
   end;

// For each cell of each partitions, count the number of loosing configurations
// Each configuration has the same unvisited cells
// Loop over the partitions
for nCL:=0 to 8 do begin
   // Loop over each cell of the configurations of this partition
   // to find a cell with the minimum probability to loose
   nCellBest:=1;  nLostBest:=30000;
   nCell:=0;
   nLost:=0;
   // anIndPart[nCL] = 1 => only one configuration in this partition, it is winning
   if (anIndPart[nCL] >= 2) then repeat
      // Loop over the configurations of the partition nCL
      inc(nCell);
      nLost:=0;
      for nInd:=1 to anIndPart[nCL] do begin
         // Count the number of loosing configurations
         // if the cell  "nCell"  is clicked on
         if (arConfig[aanPart[nCL,nInd]].anC[anCellFreeNew[nCell]] >= 09) then begin
            inc(nLost);
            end;
         end; // for nInd:=1 to anIndPart[nCL] do begin

      if (nLost = 0) and
         (anIndPart[nCL] < nNbConfList) then begin
         // The cell  "nCell"  has no bombs for all configurations of the
         // partition  nCL and
         // clicking on it will split the partition into smaller ones.
         // We can click on this cell with security

//DBW(1, 'nLost=0, nbPart=' + ITS5(anIndPart[nCL]) + ITS5(nCell));

         // Recursive call to count the number of winning games.
         nNbLostGame:=nNbLostGame + NbLostConfig(
                       anCellFreeNew[nCell], aanPart[nCL], anIndPart[nCL],
                       anCellFreeNew,
                       nNbCellFree-1,
                       nNbLostConfigMax - nNbLostGame);
         end // if (nLost = 0) then ...
      else begin
         // Memorise the cell having the less config with bombs
         if (nLostBest > nLost) then begin
            nLostBest:= nLost;
            nCellBest:= nCell;
            end;

not tested-----------------
         // Memorise the number of loosing config.
         anLostCell[nCell]:=nLost;
----------------------------------
         end; // if (nLost = 0) then ...

      until (nCell >= nNbCellFree-1) or (nLost = 0);

   if (nLost <> 0) then begin
      // No cell has been found where it can be clicked with 100% security.
      // Here we can either click on a cell having a lower probability to have a bomb
      // or we can search which cell wil maximise the probability to win.
      // This second option is better, but time consuming.

//DBW(1024*8, 'nLostBest, nbPart=' + ITS5(nLostBest) + ITS5(anIndPart[nCL]));

      // Choose the first option.
      // Recursive call to count the number of loosing games.
      nNbLostGameAdd:=NbLostConfig(
                    anCellFreeNew[nCellBest], aanPart[nCL], anIndPart[nCL],
                    anCellFreeNew,
                    nNbCellFree-1,
                    nNbLostConfigMax - nNbLostGame);

      nNbLostGame:=nNbLostGame + nNbLostGameAdd;

not tested----------------------------------
      // we should do better like this.
useless here      // 1) Sort  anLostCell  by increasing values
      // 2) see which one is better to use :
      nNbLostGameAddBest:=nNbLostConfigMax - nNbLostGame;

      for nCell:=1 to nNbCellFree-1 do begin
         // We can try to restrict to only the "best" cells
         if (anLostCell[nCell] = nLostBest) then begin
            nNbLostGameAdd:=NbLostConfig(
                    anCellFreeNew[nCell], aanPart[nCL], anIndPart[nCL],
                    anCellFreeNew,
                    nNbCellFree-1,
                    nNbLostConfigMax - nNbLostGame);

            if (nNbLostGameAddBest > nNbLostGameAdd) then begin
               // Store the result of the best choice
               nNbLostGameAddBest := nNbLostGameAdd;
               end;
            end;
         end; // for

      nNbLostGame:=nNbLostGame + nNbLostGameAddBest;
-------------------------------------------------

      end; // if (nLost <> 0) then begin

   if (nNbLostGame >= nNbLostConfigMax) then begin
      // bad cell to click on.
      NbLostConfig:=nNbLostGame;
      exit;
      end;

   end; // for nCL:=0 to 8 do begin

NbLostConfig:=nNbLostGame;
end; // TMineSolver.NbLostConfig

function TMineSolver.FindBestCell:integer;
//========================================
// When every cell has a chance to have a bomb, try to find
// the cell which maximise the probability to win the game.
// It is not always the cell which minimise the probability to have a bomb.
// Return the cell maximising the change to win
// Cell = (abXFree[return], abYFree[return])
var
  nConf,           // To list the possible configurations index
  nCell_0,         // To loop among all cells.
  nCellBest,       // Best cell found to max. the wining prob.
  nNbLostGameBest, // number of winning games for this cell.
  nNbLostGame,     // number of winning games when clicking on nCell_0
  nIndMin,         // To sort the cells
  nInd,
  nn,ii
                   : integer;
  vMin             : Float; // to sort the cells
  anConfig         : array[0..MaxConfigStore] of integer;
  anCellFree       : array[0..MaxFreeCellStore] of integer;

begin
nNbConfig:=0;
ProbabilityExplore(11);

nCellBest:=1;
nNbLostGameBest:=nNbConfig;

for nConf:=1 to nNbConfig do anConfig[nConf]:=nConf;
for nCell_0:=1 to nNbFreeCells do anCellFree[nCell_0]:=nCell_0;

// Sort to try the cells having lower prob. to loose first
for nn:=1 to nNbFreeCells-1 do begin
   // Search the min among cells from  nn to nNbFreeCells
   vMin:=avNbWithFlag[anCellFree[nn]];
   nIndMin:=nn;
   for ii:=nn+1 to nNbFreeCells do begin
      if (vMin > avNbWithFlag[anCellFree[ii]]) then begin
         nIndMin:=ii;
         vMin:= avNbWithFlag[anCellFree[ii]];
         end;
      end; // for ii

   // Exchange cell "nn" with nIndMin
   if (nn <> nIndMin) then begin
      ii                 :=anCellFree[nn];
      anCellFree[nn]     :=anCellFree[nIndMin];
      anCellFree[nIndMin]:=ii;
//DBW(1, 'ncel=' + ITS5(anCellFree[nn]) + FTS(avNbWithFlag[anCellFree[nn]]));
      end;
   end; // for nn:=1 to nNbFreeCells-1 do begin

// Loop among all cell, to see which one will minimise the number of loosing games
for nCell_0:=1 to nNbFreeCells do
 if (avNbWithFlag[anCellFree[nCell_0]] < nNbLostGameBest) then begin
//DBW(1, 'search cell' + ITS5(anCellFree[nCell_0]));
   nNbLostGame:=NbLostConfig( anCellFree[nCell_0], anConfig, nNbConfig,
                    anCellFree, nNbFreeCells, nNbLostGameBest);

   if (nNbLostGameBest > nNbLostGame) then begin
      nNbLostGameBest:=nNbLostGame;
      nCellBest:= anCellFree[nCell_0];
      end;

//DBW(1, 'Cell =' + ITS5(abXFree[nCell_0]) + ITS5(abYFree[nCell_0]) +
//                       FTS(nNbLostGame / Float(nNbConfig)));
   end; // for nCell_0:=1 to nNbFreeCells do ...

// The best cell to click on is  abXFree[nCellBest], abYFree[nCellBest]

//DBW(1, 'Best Cell =' + ITS5(abXFree[nCellBest]) + ITS5(abYFree[nCellBest]) +
//                       FTS(nNbLostGameBest / Float(nNbConfig)));

// To verify if the predicted probability to loose is correcte statistically
if (vProbFindBest = 0) then begin
   inc(anSumFindBest[0]); // number of times this function was called
   vProbFindBest :=nNbLostGameBest / Float(nNbConfig);
   nInd:=1 + Trunc(10*vProbFindBest);
   inc(anSumFindBest[nInd]);
   end;

FindBestCell:=nCellBest;
end;  // TMineSolver.FindBestCell
}
function TMineSolver.ProbabilityExplore(wFlagsP:word32):integer;
//==============================================================
// Explore the cells to find which has the lower probabilities of
// having a bomb. Take into account the number of bombs.
// In some case, this probability will be zero, and thus
// it is sure that a cell without bomb will have been found.
var
  nCount,
  nn,         // To get at random an unvisited cell
  nRep,       // Result from SetFlag
  ii,
  nXL,nYL     : integer;

  anLimit : array[1..200] of integer;
  // All points during the backtracking where a decision has been made.

  nIndLimit : integer;
  // Index in the  anLimit  array

  avNbConfigWI : array[0..500] of Float;  // WI = Without Info.
  // Number of configuration having  k  bombs among the cells having no Info.
  // A cell has no Info, if it has no visited neighbour without bombs.

  vNbConf : Float;
  // Number of configurations found, during one trial.

  nNbCellWI : integer;
  // Number of cells witout information.

  nNbConfigSum : int32;
  // Number of time a config has been found for cells with info.

  nSetFlag    : integer;
  // Determine it a  cteFlag  or a  cteNoBomb  is set to a cell
  // 1 => cteFlag,  0 => cteNoBomb

begin
if (nNbFreeCells = 0) then begin
   // Noting to do, the cell field is full
   ProbabilityExplore:=0;
   exit;
   end;

SortFreeCells;

// If we want to store all config, use also cells Without Information
//if (wFlagsP >= 10) then nNbFreeCellsInfo:=nNbFreeCells;

if (nNbFreeCellsInfo = 0) then begin
   // vTotalConfig = number of ways to place  nNbBombs  in  nNbFreeCells  cells
   // Compute : vTotalConfig := nNbFreeCells! / ((nNbFreeCells - nNbBombs)! * nNbBombs!);
   vTotalConfig :=1;
   for nn:=nNbFreeCells - nNbBombs + 1 to nNbFreeCells do vTotalConfig:=vTotalConfig * nn;
   for nn:=2 to nNbBombs do vTotalConfig:=vTotalConfig / nn;

   // All cell has the same probability to have a bomb.
   // vProbability:=nNbBombs / Float(nNbFreeCells);
   vPrCellWIWithBomb :=vTotalConfig * nNbBombs;
   ProbabilityExplore:=0;
   exit;
   end;

aanBObis:=aanBO;
aanTSbis:=aanTS;

// Number of cells without information.
nNbCellWI:=nNbFreeCells - nNbFreeCellsInfo;

// Initialisation of the number of configuration of cells without info,
// having  nn  bombs among them.
nNbBombsBis:=nNbBombs;
if (nNbBombsBis > nNbCellWI) then nNbBombsBis:=nNbCellWI;
if (nNbBombsBis > 500) then nNbBombsBis:=500;  // Just to avoid a crash in stupid cases
avNbConfigWI[0]:=1.0;  // only one config having no bombs at all.
for nn:=1 to nNbBombsBis do
   avNbConfigWI[nn]:=avNbConfigWI[nn-1]*(1+nNbCellWI - nn) / nn;

// Initialisation of the counter for each cell of how many configuration
// exist with a bomb in the cell
for nn:=1 to nNbFreeCellsInfo do avNbWithFlag[nn]:=0;
vTotalConfig:=1e-9; // to avoid dividing by 0

// To compute the probability that a cell without info has a bomb.
// It is the same for all cells Without Info
vPrCellWIWithBomb:=0;

// Will be used to limit the number of tries to find configurations.
// Usefull if too many configurations are possible
nNbConfigSum:=0;

// Start with the first unvisited cell of the list.
nXL:=abXFree[1];
nYL:=abYFree[1];

// At the begining, we know the real number of free cells and free bombs.
nNbFreeCellsBis:=nNbFreeCellsInfo;
nNbBombsBis:=nNbBombs;

// Starting of the backtracking
nSetFlag:=1;
nIndModif:=0;
nIndLimit:=1;
anLimit[nIndLimit]:=nIndModif+1; // Mark the first level
// Begin backtracking loop
//"""""""""""""""""""""""""
repeat
   SaveValue(@nSetFlag);  // to memorise if we set cteFlag or cteNoBomb at starting
   SaveValue(@nXL);
   SaveValue(@nYL);
   SaveValue(@nNbFreeCellsBis);
   SaveValue(@nNbBombsBis);
   if (nSetFlag=1) then nRep:=SetFlagTo( nXL,nYL)  // mark has a bomb
   else                 nRep:=SetFlagTo(-nXL,nYL); // mark has no bombs

   if (nNbBombsBis < 0) then nRep :=-1; // impossible config.

   // if (no more free cells) and (nRep >= 0) and (nbBombs are in the cells without info) then begin
   if (nNbFreeCellsBis <= 0) and (nRep >= 0) and
      (nNbBombsBis <= nNbCellWI) then begin
      // More possible configurations of bombs has been found.
      // Number of bombs left among the cells without info.
      vNbConf:=avNbConfigWI[nNbBombsBis];
      inc(nNbConfigSum);

      // Memorise the results
      vTotalConfig:=vTotalConfig + vNbConf;

//DBW (64, 'Total config=' + ITS5(nNbConfigSum) + ITS5(nNbBombsBis));
      // count for each cell the number of configurations
      // with  nNbBombsBis  among the cells witout info and
      // with a bomb in the cell.
      for ii:=1 to nNbFreeCellsInfo do begin
         if (aanBObis[abXFree[ii],abYFree[ii]] <= cteFlag) then
            avNbWithFlag[ii]:=avNbWithFlag[ii] + vNbConf;
         end;

     // To compute the probability that a cell without info has a bomb.
     vPrCellWIWithBomb:=vPrCellWIWithBomb + nNbBombsBis*vNbConf;

//     if (wFlagsP >= 10) then begin
//        // Store the possible configurations in arConfig[1..nNbConfig];
//        ConfigStore(Round(vNbConf));
//        end;
     end;
    // if (nRep = -1) then Contradiction found, thus this case is not taken into account

   if (nRep >= 0) and (nNbFreeCellsBis > 0) then begin
      // Take the next free cell
      ii:=0;
      repeat
         inc(ii);
         until (ii > nNbFreeCellsInfo) or
               (aanBObis[abXFree[ii],abYFree[ii]] >= cteToSee-9);

      if (ii > nNbFreeCellsInfo) then begin
         // This error never occured
         // Error, this should have been taken into account
         nRep:=-1;
ErrorWrite(2738, 'in ProbabilityExplore');
ErrorComment(2738, 'ERROR ***** ii=' + ITS5(ii) + ITS5(nNbFreeCellsInfo));
         end
      else begin
         // Give a state to the next unvisited cell
         inc(nIndLimit);   // Memorise the state of the modifications
         anLimit[nIndLimit]:=nIndModif+1;
         nXL:=abXFree[ii];
         nYL:=abYFree[ii];
         nSetFlag:=ord(nNbBombsBis > 0); // If no more bombs, only cells without bombs are possible
         // Continue the exploration setting this cell to have a bomb (= a Flag)
         // or no bombs, if no more bombs left
         end;
      end
   else nRep:=-1;  // indicate that has to come back in the backtracking

   if (nRep < 0) then begin
      // Contradiction found, or finished
      // go back in the backtracking
      //"""""""""""""""""""""""""""""
      // Set back their original values to aanBObis and aabTObis
      // Also to the number of bombs and free cells
      repeat
         while (nIndModif > anLimit[nIndLimit]) do begin
            apnModif[nIndModif]^:=anModifValue[nIndModif];
            dec(nIndModif);
            end;

         nSetFlag:=anModifValue[nIndModif];
         dec(nIndModif);
         if (nSetFlag = 0) then begin
            // Has to do more Backtracking
            dec(nIndLimit);
            end;
         until (nSetFlag <> 0) or (nIndLimit <= 0);

      if (nSetFlag = 1) then begin
         // Try setting this cell to no bomb.
         nSetFlag:=0;
         nXL:=anModifValue[nIndModif+2];
         nYL:=anModifValue[nIndModif+3];
//            inc(nIndLimit);   // Memorise the state of the modifications
//            anLimit[nIndLimit]:=nIndModif+1;  done already
         end
      else if (nIndLimit = 0) then nIndLimit:=-1; // quit the loop
      end;
   until (nIndLimit < 0) or (nNbConfigSum > 80000) or
         (GetAsyncKeyState(VK_SHIFT) < 0);

// Here we have in  avNbWithFlag
// the number of configurations with bombs
// and in  vTotalConfig  the total number of possible configurations

// Click on cells without bomb for sure
nCount:=0;
ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);

wFlagsP:=wFlagsP mod 10;

if (nNbFreeCellsInfo < nNbFreeCells) and (vPrCellWIWithBomb = 0) then begin
   // Cells without info has no bombs for sure. Click on them
   for ii:=nNbFreeCellsInfo+1 to nNbFreeCells do begin
      if (wFlagsP = 0) then begin
          MoveMouseTo(abXFree[ii], abYFree[ii]);
          SimulateLeftClick;
          end
      else if (wFlagsP = 2) then begin
         MarkCell(abXFree[ii], abYFree[ii], RGB(0,0,0));
         end;

      inc(nCount);
      end;
   end; // if (nNbFreeCellsInfo < nNbFreeCells) and (vPrCellWIWithBomb = 0) then begin

for ii:=1 to nNbFreeCellsInfo do begin
   if (avNbWithFlag[ii] = 0) then begin
      // This cell has no bomb for sure
      if (wFlagsP = 0) then begin
          MoveMouseTo(abXFree[ii], abYFree[ii]);
          SimulateLeftClick;
          end
      else if (wFlagsP = 2) then begin
         MarkCell(abXFree[ii], abYFree[ii], RGB(0,0,0));
         end;

      inc(nCount);
      end;
   end;

// If (nCount = 0) then no cells without bomb for sure has been found.
// ProbabilityExploreMore  will have to be used to click on a cell
// having some chances to have a bomb.

if (nCount > 0) then UpdateMineSweeperContent;
ProbabilityExplore:=nCount;
end;  // TMineSolver.ProbabilityExplore

function TMineSolver.ProbabilityExploreMore(wFlagsP:word32):integer;
//==================================================================
// No cell without bomb can be found with certitude, thus
// one cell has to be clicked on, with some chances that a bomb is in it.
// Return 1 normally.
// Return 0 if no cells have been clicked. Used in tests.
// wFlagsP = 0 => normal playing.
// wFlagsP = 2 => mark the cell where the program would click.
var
  ii,nXL, //nYL,
  nXClick,nYClick, // cell on which we click
  nNbState,   // number of possible values in the cell
  nn          : integer;

  vProbability : Float;
  // To give the probability information of being a cell with bomb

begin
// Search the cell having the lowest probability of having a bomb.
nn:=1;
for ii:=2 to nNbFreeCellsInfo do
    if (avNbWithFlag[nn] > avNbWithFlag[ii]) then nn:= ii;

// By default, click on a cell having lower probability of having a bomb.
nXClick :=abXFree[nn];
nYClick :=abYFree[nn];
vProbability:=avNbWithFlag[nn] / vTotalConfig;

if (nNbFreeCellsInfo < nNbFreeCells) then begin
   // Test if clicking on a cell without information is better
   if (avNbWithFlag[nn] > vPrCellWIWithBomb / Float(nNbFreeCells - nNbFreeCellsInfo))
      or (nNbFreeCellsInfo = 0) then begin
      // Cells without info have lower probabilities of having a bomb
      nn:=nNbFreeCellsInfo+1; // Choose one at random
      vProbability:=vPrCellWIWithBomb / (vTotalConfig *
                                        Float(nNbFreeCells - nNbFreeCellsInfo));

      nXClick :=abXFree[nn];
      nYClick :=abYFree[nn];

      // Among the other cells, click on one having as few neighbour as possible
      // and being as near as possible from the other cells.

      // Try do do better here  ?!?
      // Does not make any difference
      if (aanBO[1,nYNumber] >= cteToSee-2) and
         (aanTS[1,nYNumber] = 3 - (aanBO[1,nYNumber]-cteToSee)) then begin
         // Cells have only flags or not visited neighbours
         nn:=0;
         nXClick:=1;  nYClick:=nYNumber;
         end
      else if (aanBO[nXNumber,1] >= cteToSee-2) and
         (aanTS[nXNumber,1] = 3 - (aanBO[nXNumber,1]-cteToSee)) then begin
         nn:=0;
         nXClick:=nXNumber;  nYClick:=1;
         end
      else if (aanBO[nXNumber,nYNumber] >= cteToSee-2) and
         (aanTS[nXNumber,nYNumber] = 3 - (aanBO[nXNumber,nYNumber]-cteToSee)) then begin
         nn:=0;
         nXClick:=nXNumber;  nYClick:=nYNumber;
         end
      else begin
         // Try on the top line
         nXL:=1;
         repeat
            inc(nXL);
            if (aanBO[nXL,1] >= cteToSee-4) and
               (aanTS[nXL,1] = 5 - (aanBO[nXL,1]-cteToSee)) then begin
               nn:=0;
               nXClick:=nXL;  nYClick:=1;
               nXL:=nXNumber;  // Quit the repeat ... until
               end;
            until (nXL >= nXNumber-1);

         if (nn > 0) then begin
            // Try on the bottom line
            nXL:=1;
            repeat
               inc(nXL);
               if (aanBO[nXL,nYNumber] >= cteToSee-4) and
                  (aanTS[nXL,nYNumber] = 5 - (aanBO[nXL,nYNumber]-cteToSee)) then begin
                  nn:=0;
                  nXClick:=nXL;  nYClick:=nYNumber;
                  nXL:=nXNumber; // Quit the repeat ... until
                  end;
               until (nXL >= nXNumber-1);
            end;
         end; // Better ?!?

      end; // Cells without info have lower probabilities of having a bomb
   end; // Test if clicking on a cell without information is better

if (nn > 0) and (nn <= nNbFreeCellsInfo) then begin
   // All this doesn't change the wining probability more than 0.005.
   // Among the cells having a low probability of having a bomb,
   // click on one having as few unvisited neigbours as possible.
   nNbState:=aanTS[nXClick,nYClick] + cteToSee - aanBO[nXClick,nYClick];

   for ii:=2 to nNbFreeCellsInfo do
    if (avNbWithFlag[ii] = avNbWithFlag[nn]) then begin
      // Test if Cell "ii" has less possible values
      if (aanTS[abXFree[ii],abYFree[ii]] + cteToSee -
          aanBO[abXFree[ii],abYFree[ii]] < nNbState) then begin

      // Try to have as few neighbours as possible
//      if (aanTS[abXFree[ii],abYFree[ii]] < aanTS[nXClick,nYClick]) then begin
         nn:=ii;
         nXClick :=abXFree[ii];
         nYClick :=abYFree[ii];
         nNbState:=aanTS[nXClick,nYClick] + cteToSee - aanBO[nXClick,nYClick];
         end;
      end;
   end // Click on cell with few unsvisited neigbours
else if (nn > nNbFreeCellsInfo) then begin
   // Come here in less than 1% of the games.
   nNbState:=aanTS[nXClick,nYClick] + cteToSee - aanBO[nXClick,nYClick];

   for ii:=nNbFreeCellsInfo+1 to nNbFreeCells do begin
      // Test if Cell "ii" has less possible values
      if (aanTS[abXFree[ii],abYFree[ii]] + cteToSee -
          aanBO[abXFree[ii],abYFree[ii]] < nNbState) then begin

      // Try to have as few neighbours as possible
//      if (aanTS[abXFree[ii],abYFree[ii]] < aanTS[nXClick,nYClick]) then begin
         nn:=ii;
         nXClick :=abXFree[ii];
         nYClick :=abYFree[ii];
         nNbState:=aanTS[nXClick,nYClick] + cteToSee - aanBO[nXClick,nYClick];
         end;
      end;
   end;

{ --> Makes nearly no differences !
// There are not too much configurations, see which cell
// maximises the winning chances
if (vTotalConfig < 5000) and (vTotalConfig > 2.5) and
   (nNbFreeCells <= MaxFreeCellStore) then begin
   // To store all possible configurations.
   nn:=FindBestCell;
   nXClick :=abXFree[nn];
   nYClick :=abYFree[nn];
   vProbability:=avNbWithFlag[nn] / vTotalConfig;

   // to count how many games had few config.
   fHadFewConfig:=TRUE;
   end;
}

if (nWinProbabilityExploreMore1 >= 0) and (nWinProbabilityExploreMore1 <= 15) then begin
   // Indicate that the cell has been clicked with a prob. to explode
   WriteWin( nWinProbabilityExploreMore1,
             GetMsg(10651) + IntToStr(nNbGames, 0) +
             GetMsg(10652) + IntToStr(nXClick,0) + ' ' + IntToStr(nYClick, 0));
   WriteWin( nWinProbabilityExploreMore1,
             GetMsg(10653) + FloatToStr(vProbability,9,6) +
             GetMsg(10654) + IntToStr(nXNumber*nYNumber - nNbFreeCells, 0));
   end;

if (wFlagsP = 0) then begin
   MoveMouseTo(nXClick, nYClick);
   SimulateLeftClick;
   end
else if (wFlagsP = 2) then begin
   MarkCell(nXClick, nYClick, RGB(128,0,0));
   WriteWin(0, ''); // skip a line
   WriteWin(0, FloatToStr(vProbability,9,6) + GetMsg(12124) +
               IntToStr(nXClick, 0) + ' ' + IntToStr(nYClick, 0));
   end;

vProbProduct:=vProbProduct * (1.0 - vProbability);

UpdateMineSweeperContent;
ProbabilityExploreMore:=1;
end;  // TMineSolver.ProbabilityExploreMore

procedure TMineSolver.SolveMineSweeper;
//=====================================
// Go and click to solve the minesweeper
// This will be better done in  tEx064.pp
var
  nSecurity   : integer;
begin
// Screen position of the cells
ptTopLeftScreen:=ptTopLeft;
ClientToScreen(hWinSec, ptTopLeftScreen);
SetCursorPos(ptTopLeftScreen.x+50, ptTopLeftScreen.y-40);

// A first click can be usefull
MoveMouseTo(1,1);
SetCursorPos(ptTopLeftScreen.x-4, ptTopLeftScreen.y-4);
SimulateLeftClick;

// Get the cells content.
GetMineSweeperContent(0);
fQuitSolve:=FALSE;
nNbBombExplosed:=0;

if (nNbFreeCells <= 0) then exit;

// Product of the probabilities of clicking on a bomb.
vProbProduct:=1.0;

//fHadFewConfig:=FALSE;

// If field empty, start exploring.
if (nNbFreeCells = nXNumber*nYNumber) then begin
   StartingExplore;
   if (nNbBombExplosed > 0) then exit;
   end;

// Loop to solve
nSecurity:=500;
repeat
   // Try to put flags on cells having bombs
   PutFlags(0);   if fQuitSolve then exit;

   // Try to desactivate bombs. If none is found, choose one cell at random
   if (SaftyExplore(0) = 0) and (not fQuitSolve) then begin
      // Try a more sophisticated way to find bombs
      if (nNbFreeCells <= 0) then exit; // finished

      if (nExitSolver = 1) then exit;
//DBW (8, 'Explore more');
      if (nExitSolver  = 101) then begin
         if (SaftyExploreMore(1) > 0) then exit;
         end;

      if (SaftyExploreMore(0) = 0) and (not fQuitSolve) then begin
         if (nNbFreeCells <= 0) then exit; // finished
//DBW (8, 'Probability Explore');

         if (nExitSolver = 2) then exit;
         if (nExitSolver  = 102) then begin
            // exit if no cells without bombs can be found by deduction,
            // but cells without bombs can be found by probabilities.
            if (ProbabilityExplore(1) > 0) then exit;
            end;

         // Try solving using probabilities
         if (ProbabilityExplore(0) = 0) and (not fQuitSolve) then begin
            if (nNbFreeCells <= 0) then exit; // finished
            if (nExitSolver = 3) then exit;

            if (nExitSolver = 4) and (vTotalConfig < 1000) and
               (vTotalConfig > 6) then exit;

            if (nExitSolver = 5) and (vTotalConfig < 100) and
               (vTotalConfig > 6) then exit;

            // No cells without bombs for sure have been found, thus
            // click on one cell with some chances to click on a bomb.
//DBW (8, 'Probability Explore with explosion risks');
            if (ProbabilityExploreMore(0) = 0) then exit;

            end; // if (ProbabilityExplore = 0)
         end; // if (SaftyExploreMore = 0) then begin
      end; // if (SaftyExplore = 0) then begin

   dec(nSecurity);
   until (nNbBombExplosed > 0) or (nNbFreeCells <= 0) or (nSecurity <= 0) or
         fQuitSolve;
         
if (nSecurity <= 0) then begin
   DBW(1, 'nSecurity <= 0,  nNbFreeCells =' + ITS5(nNbFreeCells));
   nNbBombExplosed:=5;
   end;
   
end;  // TMineSolver.SolveMineSweeper

procedure TMineSolver.GetStatistiquePlay;
//=======================================
// Plays many Minesweeper games to get statistics on it.
var
  nNbWin,    // number of times we win
  nn
             : int32;
  ch         : char;
  vMeanProb  : Float;
  anNbBombEnd : array[0..500] of int32;
  // count the number of bombs which exploded at the end of game
  vSumProb    : Float;

begin
for nn:=0 to 500 do anNbBombEnd[nn]:=0; // no games yet => no bombs

//nNbFewConfigWin:=0;
//nNbFewConfigLost:=0;

// To display some statistics
if (nWinGetStatistiquePlay1 >= 0) and (nWinGetStatistiquePlay1 <= 15) then begin
   WriteWin( nWinGetStatistiquePlay1, GetMsg(10700));
   end;

nNbGames :=0;
nNbWin   :=0;
vMeanProb:=0;
repeat
   // To slow down the mouse speed
   if (GetAsyncKeyState(VK_S) < 0) then begin
      dec(nSpeed);
      if (nSpeed <= 0) then nSpeed:=0;
      fIsKeyUP:=FALSE;
      end;

   inc(nNbGames);
   ClickOnNewGame;
   SolveMineSweeper;

   if not fQuitSolve then begin
      vMeanProb:=vMeanProb + vProbProduct;

      // Count how many games ended with this number of bombs
      inc(anNbBombEnd[nNbBombExplosed]);

      if (nNbBombExplosed = 0) then begin
          // Game won
          inc(nNbWin);
          ch:='*';
          vMeanProb:=vMeanProb + vProbProduct;
          end
      else begin
         // Game lost
         ch:=' ';
         if      (nSpeed <= 2) then DBWaitSec(3) // wait 3 seconds to let us see
         else if (nSpeed  = 3) then DBWaitSec(2) // wait 2 seconds to let us see
         else if (nSpeed <= 5) then DBWaitSec(1); // wait 1 seconds to let us see
         end;

//      if (fHadFewConfig) then begin
//         if (nNbBombExplosed = 0) then inc(nNbFewConfigWin)
//         else                          inc(nNbFewConfigLost);
//         end;

      if (nWinGetStatistiquePlay1 >= 0) and (nWinGetStatistiquePlay1 <= 15) and
         (nDisplayFrequ > 0) and (nNbGames mod nDisplayFrequ = 0) then begin
         WriteWin( nWinGetStatistiquePlay1, GetMsg(10701) +
                   ITS5(nNbGames) + ch + FTS(nNbWin / Float(nNbGames)) +
                   ITS5(nNbBombExplosed));
          end;
      end // if not fQuitSolve then begin
   else dec(nNbGames); // last game has been played till the end

   until (GetAsyncKeyState(VK_CONTROL) < 0) or
         ((nNbBombExplosed = 0) and (nNbFreeCells > 0));
   // The last condition is satisfied if a stopping condition has been set
   // and has been reached.

DBW(1, 'fQuitSolve=' + ITS5(integer(fQuitSolve)));
DBW(1, 'keyState=' + ITS(integer(GetAsyncKeyState(VK_CONTROL))));
   
if (nNbGames = 0) then exit;

// Show statistics of how many bombs exploded per games
if (nWinGetStatistiquePlay2 >= 0) and (nWinGetStatistiquePlay2 <= 15) then begin
   WriteWin( nWinGetStatistiquePlay2, GetMsg(10710));
   WriteWin( nWinGetStatistiquePlay2, GetMsg(10711));
   WriteWin( nWinGetStatistiquePlay2,
             IntToStr(nNbGames,0) + GetMsg(10712));
   WriteWin( nWinGetStatistiquePlay2, GetMsg(10713));
   vSumProb:=0;
   nn:=0;
   repeat
      vSumProb:=vSumProb + anNbBombEnd[nn] / Float(nNbGames);
      WriteWin( nWinGetStatistiquePlay2,
                   IntToStr(nn, 8) + IntToStr(anNbBombEnd[nn],10) +
                   FloatToStr(anNbBombEnd[nn] / Float(nNbGames),11,6) +
                   FloatToStr(vSumProb,10,6));
      inc(nn);
      until (nn > 500) or (vSumProb > 0.999);
   end;
end;  // TMineSolver.GetStatistiquePlay

procedure TMineSolver.HintAreFlagCorrect;
//=======================================
// Determine if the flags are set correctly
var
  nX,nY,
  nWrongPlaced, // number of flags placed on cells which are not sure to have a bomb
  nMissing,     // number of missing flags
  nSetByDeduction // Number of flags set by more complicate deduction
                : integer;
  aanBOLoc : array[0..36,0..31] of integer;

begin
// Get the actual content of the game and memorise it.
GetMineSweeperContent(0);
aanBOLoc:=aanBO;

// Get the actual content of the game setting the cells with flags
// and questions ToSee
GetMineSweeperContent(6);

// Determine the cells which need a Flag.
PutFlags(1);

// search using more deductions.
nSetByDeduction:=PutFlagsMore(1);

// See the difference between before and after counting the right flags
nMissing:=0;
nWrongPlaced:=0;
for nX:=1 to nXNumber do for nY:=1 to nYNumber do begin
   if (aanBO[nX,nY]    >= cteFlag-9) and (aanBO[nX,nY]    <= cteFlag) and
      (aanBOLoc[nX,nY] > cteFlag) then begin
      // A flag should be placed, but none are there
      inc(nMissing);
      end;

   if (aanBOLoc[nX,nY] >= cteFlag-9) and (aanBOLoc[nX,nY] <= cteFlag) and
      (aanBO[nX,nY]    > cteFlag) then begin
      // A flag is placed, but without being sure that it is right
      inc(nWrongPlaced);
      end;
   end;

WriteWin(0, ''); // skip a line
if (nMissing = 0) and (nWrongPlaced = 0) then WriteWin(0, GetMsg(12010))
else begin
   // Flags are not optimally placed
   if (nMissing = 0)     then WriteWin(0, GetMsg(12011))
   else WriteWin(0, IntToStr(nMissing,0) + GetMsg(12013));
   if (nWrongPlaced = 0) then WriteWin(0, GetMsg(12012))
   else WriteWin(0, IntToStr(nWrongPlaced,0) + GetMsg(12014));

   if (nSetByDeduction > 0) then
      WriteWin(0, IntToStr(nSetByDeduction,0) + GetMsg(12015));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintAreFlagCorrect

procedure TMineSolver.HintHasBomb;
//================================
// Count the number of cells which can be have a bomb and are not
// already marked
var
  nSetByDeduction, // Number of flags set by more complicate deduction
  nNbWithBombs : integer;
begin
GetMineSweeperContent(0);
nNbWithBombs:=PutFlags(1);

// search using more deductions.
nSetByDeduction:=PutFlagsMore(1);

WriteWin(0, ''); // skip a line
WriteWin(0, IntToStr(nNbWithBombs + nSetByDeduction,3) + GetMsg(12021));

if (nSetByDeduction > 0) then
   WriteWin(0, IntToStr(nSetByDeduction,0) + GetMsg(12022));

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintHasBomb

procedure TMineSolver.HintEasillyFound;
//=====================================
// Count the number of cells which can be easilly found not having a bomb.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);

nNbWithoutBombs:=SaftyExplore(1);

WriteWin(0, ''); // skip a line
WriteWin(0, IntToStr(nNbWithoutBombs,3) + GetMsg(12031));

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintEasillyFound

procedure TMineSolver.HintDeductionFound;
//=======================================
// Count the number of cells which can be found by deduction not having a bomb.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);

nNbWithoutBombs:=SaftyExploreMore(1);

WriteWin(0, ''); // skip a line
WriteWin(0, IntToStr(nNbWithoutBombs,3) + GetMsg(12032));

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintDeductionFound

procedure TMineSolver.HintProbabilityFound;
//=========================================
// Count the number of cells which can be found having no bomb,
// using probabilities. This finds exactly all cells having no bombs for sure.
// It computes also the probability of having a bomb for each cell.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);

nNbWithoutBombs:=ProbabilityExplore(1);

WriteWin(0, ''); // skip a line
WriteWin(0, IntToStr(nNbWithoutBombs,3) + GetMsg(12033));

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintProbabilityFound

procedure TMineSolver.HintMarksHasBombEasyFound;
//==============================================
// Mark the cells which can be easilly found having a bomb.
var
  nNbWithBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);
aanBObis:=aanBO;
nNbWithBombs:=PutFlags(2);

if (nNbWithBombs = 0) then begin
   WriteWin(0, ''); // skip a line
   WriteWin(0, GetMsg(12111));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarksHasBombEasyFound

procedure TMineSolver.HintMarksHasBombDeductionFound;
//===================================================
// Mark the cells which can be found having a bomb.
var
  nSetByDeduction : integer;
begin
// Get the content.
GetMineSweeperContent(0);

// Mark the cells having bombs, using more deductions.
nSetByDeduction:=PutFlagsMore(2);

if (nSetByDeduction = 0) then begin
   WriteWin(0, ''); // skip a line
   WriteWin(0, GetMsg(12112));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarksHasBombDeductionFound

procedure TMineSolver.HintMarkEasillyFound;
//=========================================
// Mark the cells which can be easilly found not having a bomb.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);
aanBObis:=aanBO;

// Mark the cells found.
nNbWithoutBombs:=SaftyExplore(2);

if (nNbWithoutBombs = 0) then begin
   WriteWin(0, ''); // skip a line
   WriteWin(0, GetMsg(12121));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarkEasillyFound

procedure TMineSolver.HintMarkDeductionFound;
//===========================================
// Mark the cells which can be found not having a bomb, by deduction.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);
aanBObis:=aanBO;

// Mark the cells found.
nNbWithoutBombs:=SaftyExploreMore(2);

if (nNbWithoutBombs = 0) then begin
   WriteWin(0, ''); // skip a line
   WriteWin(0, GetMsg(12122));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarkDeductionFound

procedure TMineSolver.HintMarkProbabilityFound;
//=============================================
// Mark the cells which can be found having no bomb
// using probabilities.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);
aanBObis:=aanBO;

// Mark the cells found.
nNbWithoutBombs:=ProbabilityExplore(2);

if (nNbWithoutBombs = 0) then begin
   WriteWin(0, ''); // skip a line
   WriteWin(0, GetMsg(12123));
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarkProbabilityFound

procedure TMineSolver.HintMarkProbabilityTryFound;
//================================================
// Mark the cells which would be played by the program
// using probabilities.
var
  nNbWithoutBombs : integer;
begin
// Get the content.
GetMineSweeperContent(0);
//aanBObis:=aanBO;

// Mark the cells found.
nNbWithoutBombs:=ProbabilityExplore(2);

if (nNbWithoutBombs = 0) then begin
   // Mark the cell played by the program, with a risk of bomb explosion
   ProbabilityExploreMore(2);
   end;

// Get again the content, to be up to date
GetMineSweeperContent(0);
end;  // TMineSolver.HintMarkProbabilityFound

function myCreateWindowTextObj(strS:ShortString):Pointer;
//=======================================================
var poObj : PMyWindowText;
begin
new(poObj, Init(strS));
myCreateWindowTextObj:=poObj;
end;  // myCreateWindowTextObj

//******************************************************************3
//***** TMyWindowText = object(TWindowText)
//******************************************************************3

constructor TMyWindowText.Init(strWinTitlePrm:ShortString);
//=========================================================
begin
inherited Init(strWinTitlePrm);
end;  // TMyWindowText.Init

procedure TMyWindowText.WinCreate(strWinClass:ShortString);
//=========================================================
// Create the window.
begin
inherited WinCreate(strWinClass);

if not odd(wWinNumber) then begin
   // Sets courier font for even window number
   with oRichEditFrame.TheLogFont do  begin
      lfHeight         := 14;                // Default logical height of font
      lfWidth          := 12;                // Default logical average character width
      lfEscapement     := 0;                // angle of escapement
      lfOrientation    := 0;                // base-line orientation angle
      lfWeight         := FW_NORMAL;        // font weight
      lfItalic         := 0;                // italic attribute flag
      lfUnderline      := 0;                // underline attribute flag
      lfStrikeOut      := 0;                // strikeout attribute flag
      lfCharSet        := DEFAULT_CHARSET;  // character set identifier
      lfOutPrecision   := OUT_DEFAULT_PRECIS;  // output precision
      lfClipPrecision  := CLIP_DEFAULT_PRECIS; // clipping precision
      lfQuality        := DEFAULT_QUALITY;     // output quality
      lfPitchAndFamily := DEFAULT_PITCH;    // pitch and family
      Strcopy(lfFaceName,'Courier New');    // pointer to typeface name string
      end; // with

   oRichEditFrame.TheFont  := CreateFontIndirect(@oRichEditFrame.TheLogFont);
   SendMessage(oRichEditFrame.hWin, WM_SETFONT, oRichEditFrame.TheFont, 1);

   // TheFont will be destroyed automaticaly
   end;
end;  // TMyWindowText.WinCreate

//******************************************************************3
//***** TMyWindow = object(TMyWindowText)
//******************************************************************3

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
var nn : integer;
begin
inherited Init(strWinTitlePrm);

wMsgMenuNr:=10000;

for nn:=0 to 99 do apoDialogHelp[nn]:=NIL;
poDialogProb:=NIL;

// Defines that  WriteWin(0, string)  will write the string into the window
SetWriteWinNr(0, @Self);

oSolve.Init;
oSolve.poWinMain:=@Self;
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//=========================
begin
oSolve.Done;

inherited Done;
end;  // TMyWindow.Done

procedure TMyWindow.WinCreate(strWinClass:ShortString);
//=====================================================
// Create the window.
begin
inherited WinCreate(strWinClass);
end;  // TMyWindow.WinCreate

procedure TMyWindow.GetSecWindowHandle;
//=====================================
// Get the handle of the window which is being activated
begin
// By clicking on another window, normally the MineSweeper window,
// Gets the handle of that window.
oSolve.hWinSec := 0;

// These two values indicate that when the main window is desactivated,
// The new active window is the wanted one.
pDialogActivate:=PDialog(DialogCreateMSG(20199, @Self));

SetActiveWindow(hWin);
SetFocus(hWin);
end;  // TMyWindow.GetSecWindowHandle

procedure TMyWindow.ButtonClicked(poCtrl:PWindowBase0; wCtrlID:Word32);
//=====================================================================
// To react when clicking on the button in the dialog 21199
begin
poCtrl:=poCtrl; // to avoid a hint

if (wCtrlID <> 21101) then exit;

with oSolve do begin
   // Update the probabilities of having a bomb in the cells
   GetMineSweeperContent(0);
   ProbabilityExplore(1);
   end; // with oSolve do begin
end;  // TMyWindow.ButtonClicked

procedure TMyWindow.WMTimer(wPar:word32);
//=======================================
var
  ptL         : TPoint; // mouse cursor position in screen pixel coord.
  nXL,nYL,              // corresponding cell coordonate
  nn          : int32;
//  nNbFreeCellsMemL : integer;
              // Memorise the number of configurations before last update
begin
wPar:=wPar; // to avoid a hint
//messageBeep(0);

with oSolve do begin
   if (hWinSec = 0) then exit;

   // !!!!! Addition if the dialog box update regularly the probabilities.
   ButtonClicked(oSolve.poWinMain, 21101);
   
   // Get the mouse position, to get the corresponding cell coordonate.
   GetCursorPos(@ptL);

   // position of the top left cell
   ptTopLeftScreen:=ptTopLeft;
   ClientToScreen(hWinSec, ptTopLeftScreen);

   nXL:=1 + (ptL.x - ptTopLeftScreen.x) div nXDelta;
   nYL:=1 + (ptL.y - ptTopLeftScreen.y) div nYDelta;

   nXPosCell :=nXL;
   nYPosCell :=nYL;

   if (nXL < 1) or (nXL > nXNumber) or (nYL < 1) or (nYL > nYNumber) then begin
      vBombProbability:=0.0;

      // Update the content of the dialog box
      UpdateFromVar;
      exit; // not over a cell
      end;

   // Search the corresponding cell index in "avNbWithFlag"
   nn:=0;
   repeat
      inc(nn);
      if (abXFree[nn] = nXL) and (abYFree[nn] = nYL) then begin
         vBombProbability:=avNbWithFlag[nn] / vTotalConfig;

         // Update the content of the dialog box
         UpdateFromVar;

         exit;
         end;

      until (nn >= nNbFreeCellsInfo);

   // The cursor is over a cell without info.
   if (nNbFreeCells > nNbFreeCellsInfo) then
      vBombProbability:= vPrCellWIWithBomb /
                         (vTotalConfig * Float(nNbFreeCells - nNbFreeCellsInfo))
   else begin
      // Not over a free cell
      vBombProbability:=0.0;
      end;

   // Update the content of the dialog box
   UpdateFromVar;
   end; // with oSolve do begin

end;  // TMyWindow.WMTimer

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// Handles the menu messages
Var wNrMenu     : word;

begin
WMCommand := 0; // Default return value when processing a message

if (lPar <> 0) then begin
   // call the parent WMCommand method.
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;

// for the menus,  lPar = 0
//=========================
wNrMenu   := LoWord(wPar);

// Menus where the hWinSec is not obliged to exist.
with oSolve do Case wNrMenu Of
   201 : GetSecWindowHandle;

   // Options -> 500
   504 : DialogExecMSG(30299, @Self);
   506 : SelectGameVersion;
   508 : DialogExecMSG(50899, @Self); // Set the number of bombs
   510 : DialogExecMSG(51099, @Self); // Set display options.

   521 : DialogExecMSG(52199, @Self); // Set Stopping conditions

   551 : GetCellPositions;
   561 : DialogExecMSG(30199, @Self); // Sizes of the cells and minesweeper field

   // Developping tests -> 600
   604 : DialogExecMSG(60499, @Self);

   // Help information -> 900
//   901..999 : DialogExecMSG(100*wNrMenu + 99, @Self);
   901..999 : if (apoDialogHelp[wNrMenu-900] <> NIL) then begin
                 SetActiveWindow(apoDialogHelp[wNrMenu-900]^.hWin);
                 end
              else begin
                 apoDialogHelp[wNrMenu-900]:=PDialog(DialogcreateMSG(100*wNrMenu + 99, @Self));
                 end;

   else if (wNrMenu < 30000) and IsSecWindowDefined then begin
      SetForegroundWindow(hWinSec);

      Case wNrMenu Of
         // Menus where the hWinSec must exist.
         // Starting -> 200

         // Solver Hints -> 300
         301 : HintAreFlagCorrect;
         303 : HintHasBomb;
         304 : HintEasillyFound;
         305 : HintDeductionFound;
         306 : HintProbabilityFound;

         311 : HintMarksHasBombEasyFound;
         312 : HintMarksHasBombDeductionFound;
         313 : HintMarkEasillyFound;
         314 : HintMarkDeductionFound;
         315 : HintMarkProbabilityFound;
         316 : HintMarkProbabilityTryFound;

         331 : begin // Use a timer to show the bomb probabilities of the cells
               GetMineSweeperContent(0);
               ProbabilityExplore(1); // To be up to date in the probabilities of having a bomb

               if (poDialogProb <> NIL) then begin
                  // dialog already opened, activate it.
                  SetActiveWindow(poDialogProb^.hWin);
                  exit;
                  end;

               SetTimer( hWin, TIMER_ID,
                       300,  // timer frequency in mili-seconds
                       NIL); // a WM_TIMER will be send to  CtrlWindowProc

               // To start the dialog box with a content.
               if (vTotalConfig = 0) then vTotalConfig:=1; // to avoid div. by zero
               nXPosCell :=-1;
               nYPosCell :=-1;
               vBombProbability:=0;

               poDialogProb:=PDialog(DialogcreateMSG(21199, @Self));
               end; // 331

         // Solver  -> 400
         401 : begin GetMineSweeperContent(0);  PutFlags(0);  end;
         402 : begin GetMineSweeperContent(0);  PutFlagsMore(0);  end;
         403 : begin GetMineSweeperContent(0);  SaftyExplore(0);  end;
         404 : begin GetMineSweeperContent(0);  SaftyExploreMore(0);  end;
         405 : begin
               GetMineSweeperContent(0);
               if (ProbabilityExplore(0) = 0) then begin
                  // No cell can be found having NO bomb for sure.
                  ProbabilityExploreMore(0);  // click with risks of hitting a bomb.
                  end;
               end; // 404
         411 : begin
               if (poDialogProb <> NIL) then KillTimer(hWin, TIMER_ID);
               nNbGames:=0;
               SolveMineSweeper;
               if (poDialogProb <> NIL) then SetTimer( hWin, TIMER_ID, 200, NIL);
               end;

         421 : begin
               if (poDialogProb <> NIL) then KillTimer(hWin, TIMER_ID);
               GetStatistiquePlay;
               if (poDialogProb <> NIL) then SetTimer( hWin, TIMER_ID, 200, NIL);
               end;

         // Options -> 500

         // Developping tests -> 600
         601 : SimulateLeftClick;
         602 : SimulateRightClick;
         603 : GetMineSweeperContent(1);
         611 : ClickOnNewGame;

         // Help information -> 900
         else WMCommand:=inherited WMCommand(wPar, lPar);
         end; // Case wNrMenu Of
      end  // else if (wNrMenu < 30000) and IsSecWindowDefined then begin
   else WMCommand:=inherited WMCommand(wPar, lPar);
   end; // else of with oSolve do Case wNrMenu Of

SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
SetActiveWindow(hWin);
SetFocus(hWin);
SetForegroundWindow(hWin);
end;  // TMyWindow.WMCommand

procedure TMyWindow.AskAcceptDlgQuit(poDlg:PWindowBase0; wCtrlIDP:Word32; var wRep:word32);
//=========================================================================================
// notification that the dialog box pointed by  PDialog(lPar)
// is about to quit, accepting its changes (With OK).
// If we refuse to quit, just Set  wRep:=0;
begin
wCtrlIDP:=wCtrlIDP;  wRep:=wRep;  // to avoid a hint
// notification that the dialog box pointed by  pDlg is about to quit.
// Its the best place to check if we accept to quit.
if (poDlg = PWindowBase0(pDialogActivate)) then begin
   // Closing the dialog Box, sets the pointer on it to NIL.
   pDialogActivate:=NIL;
   end; // if (poDlg = pDialogActivate) then begin

if (PDialog(poDlg)^.wMsgDescriptor > 90000) then begin
   apoDialogHelp[(PDialog(poDlg)^.wMsgDescriptor div 100) - 900]:=NIL;
   end;

if (PDialog(poDlg)^.wMsgDescriptor = 21199) then begin
   // The dialog used with the timer is closed, kill the timer.
   KillTimer(hWin, TIMER_ID);
   poDialogProb:=NIL;
   end;
end;  // TMyWindow.AskAcceptDlgQuit

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//======================================================================================
var
  pppT : Point;

begin
if (wMsg = WM_ACTIVATE) and (LoWord(wPar) = WA_INACTIVE) and
   (pDialogActivate <> NIL) and
   (HWND(lPar) <> pDialogActivate^.hWin) then begin
   // Get the handle of the second window, usually the MineSweeper
//   oSolve.hWinSec:=GetForegroundWindow;
   GetCursorPos(@pppT);   
   oSolve.hWinSec:=WindowFromPoint(pppT);
   
   // Closes the dialog Box
   DestroyWindow(pDialogActivate^.hWin);
   pDialogActivate:=NIL;

   // Get the position and size of the Minesweeper, and
   // deduce from it the number of cells
   oSolve.GetSecWindowPosition;
   end;

if (wMsg = WM_Timer) then begin
   // To show the probability that a cell has a bomb
   WMTimer(wPar);
   WindowProc:=0;
   exit;
   end;

// Default treatment of a message. Does nearly nothing
WindowProc := inherited WindowProc(hWindow, wMsg, wPar, lPar);
end;  // TMyWindow.WindowProc

Begin // MAIN ******
// Assign the pointer on the function which creates
// the TWindowText object.
// This allows the library to create ojects defined by the user
// in the final application.
gglfnCreateWindowTextObj:=@myCreateWindowTextObj;

// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example065');

// Creation of the main window
new(poMyMainWin, Init('tEx065. Solving the MineSweeper game.'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

