{ Test some sound possibilities

Information to play a self defined sound.
The PlaySound function plays a sound specified by the given filename, resource, or system event. 
(A system event may be associated with a sound in the registry or in the WIN.INI file.)

BOOL PlaySound(
    LPCSTR pszSound,
    HMODULE hmod,	
    DWORD fdwSound	
   );	
Parameters
==========
pszSound
A string that specifies the sound to play. If this parameter is NULL, any currently playing waveform sound is stopped. To stop a non-waveform sound, specify SND_PURGE in the fdwSound parameter.
Three flags in fdwSound (SND_ALIAS, SND_FILENAME, and SND_RESOURCE) determine whether the name is interpreted as an alias for a system event, a filename, or a resource identifier. If none of these flags are specified, PlaySound searches the registry or the WIN.INI file for an association with the specified sound name. If an association is found, the sound event is played. If no association is found in the registry, the name is interpreted as a filename.

hmod
Handle of the executable file that contains the resource to be loaded. This parameter must be NULL unless SND_RESOURCE is specified in fdwSound.

fdwSound
Flags for playing the sound. The following values are defined:

SND_APPLICATION
The sound is played using an application-specific association.

SND_ALIAS
The pszSound parameter is a system-event alias in the registry or the WIN.INI file. Do not use with either SND_FILENAME or SND_RESOURCE.

SND_ALIAS_ID
The pszSound parameter is a predefined sound identifier.

SND_ASYNC
The sound is played asynchronously and PlaySound returns immediately after beginning the sound. To terminate an asynchronously played waveform sound, call PlaySound with pszSound set to NULL.

SND_FILENAME
The pszSound parameter is a filename.

SND_LOOP
The sound plays repeatedly until PlaySound is called again with the pszSound parameter set to NULL. You must also specify the SND_ASYNC flag to indicate an asynchronous sound event.

SND_MEMORY
A sound events file is loaded in RAM. The parameter specified by pszSound must point to an image of a sound in memory.

SND_NODEFAULT
No default sound event is used. If the sound cannot be found, PlaySound returns silently without playing the default sound.

SND_NOSTOP
The specified sound event will yield to another sound event that is already playing. If a sound cannot be played because the resource needed to generate that sound is busy playing another sound, the function immediately returns FALSE without playing the requested sound.
If this flag is not specified, PlaySound attempts to stop the currently playing sound so that the device can be used to play the new sound.

SND_NOWAIT
If the driver is busy, return immediately without playing the sound.

SND_PURGE
Sounds are to be stopped for the calling task. If pszSound is not NULL, all instances of the specified sound are stopped. If pszSound is NULL, all sounds that are playing on behalf of the calling task are stopped.
You must also specify the instance handle to stop SND_RESOURCE events.

SND_RESOURCE
The pszSound parameter is a resource identifier; hmod must identify the instance that contains the resource.

SND_SYNC
Synchronous playback of a sound event. PlaySound returns after the sound event completes.

Return Values
Returns TRUE if successful or FALSE otherwise.

Remarks
The sound specified by pszSound must fit into available physical memory and 
be playable by an installed waveform-audio device driver. PlaySound searches 
the following directories for sound files: the current directory; 
the Windows directory; the Windows system directory; directories listed in
the PATH environment variable; and the list of directories mapped in a network.
For more information about the directory search order, see the documentation for the OpenFile function.

If it cannot find the specified sound, PlaySound uses the default system event sound entry instead. 
If the function can find neither the system default entry nor the default sound, 
it makes no sound and returns FALSE.
  --------------------------------------------}
program tEx072_Sound_Tests;

// To tell FreePascal that the application is a Graphical User Interface
// i.e. is running in windows, not in text mode.
{$APPTYPE GUI}

Uses
  wApplication0
 ,wWindowText0
 ,wWindow0
 ,wDialog0
 ,wPopupWindow0
 ,wDialogCreatorLOW2
 ,wDialogCreator2
 ,wVariableNumLink2
 ,wWindowBase0
 ,wErrors0
 ,wGlobals2
 ,wDebug0
 ,Windows
 ;
// cette partie est une rcupration des constantes de playsound,
// j'ai laiss les commentaires originaux qui peuvent etre utiles
const
SND_SYNC = $0000; // play synchronously (default)
SND_ASYNC = $0001; // play asynchronously
SND_NODEFAULT = $0002; // don't use default sound
SND_MEMORY = $0004; // lpszSoundName points to a memory file
SND_LOOP = $0008; // loop the sound until next sndPlaySound
SND_NOSTOP = $0010; // don't stop any currently playing sound
SND_NOWAIT = $00002000; // don't wait if the driver is busy
SND_ALIAS = $00010000; // name is a registry alias
SND_ALIAS_ID = $00110000; // alias is a predefined ID
SND_FILENAME = $00020000; // name is file name
SND_RESOURCE = $00040004; // name is resource name or atom
SND_PURGE = $0040; // purge non-static events for task
SND_APPLICATION = $0080; // look for application specific association

// declaration de la fonction playsound contenue dans la dll winmm.dll
function PlaySound(pszSound: PChar; hmod: HMODULE; fdwSound: DWORD): BOOL;
stdcall; external 'winmm.dll' name {$ifdef unicode}'PlaySoundW' {$else}
'PlaySoundA' {$endif};


type
  PMyDialog = ^TMyDialog;
  TMyDialog = object(TDialog)
  //"""""""""""""""""""""""""
  constructor Init(poParentW:PWindowBase);

  function WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL; virtual;
  // Handles the menu messages

//  function DialogProc(hDialog:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
    // Not used
  // Handles the dialog messages. See comments in  wDialog0.pp
  end; // TMyDialog = object(TDialog)

type
  PMyWindow = ^TMyWindow;
  TMyWindow = object(TWindow)
  //"""""""""""""""""""""""""
  constructor Init(strWinTitlePrm:ShortString);

  destructor Done; virtual;

  procedure WinCreate(strWinClass:ShortString); virtual;
  // When creating the window, we ask to show the dialog box, not the window

  function WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  // Handles the menu messages

  function WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT; virtual;
  end; // TMyWindow = object(TWindow)

var
  oApp        : TApplication;
  poMyMainWin : PMyWindow;

function myCreateDialogObj(poParent:Pointer):Pointer;
//===================================================
var poObj : PMyDialog;
begin
new(poObj, Init(poParent));
myCreateDialogObj:=poObj;
end;  // myCreateDialogObj

//******************************************************************
// TMyDialog = object(TDialog)
//******************************************************************

constructor TMyDialog.Init(poParentW:PWindowBase);
//================================================
begin
// must be done, even if it seems useless.
inherited Init(poParentW);
end;  // TMyDialog.Init

function TMyDialog.WMCommand(wPar:WPARAM; lPar:LPARAM):BOOL;
//==========================================================
// Handles the menu messages
var
  wCtrlID     : word32;

begin
if (HiWord(wPar) <> BN_CLICKED) then begin
   WMCommand:=inherited WMCommand(wPar, lPar);
   exit;
   end;
//DBW(1, 'WMCommand=' + DBNotifToStr(HiWord(wPar)) + ITS5(LoWord(wPar)));
wCtrlID:=LoWord(wPar);

//MessageBeep(0);

//DBW(1, 'Button clicked=' + ITS(poBtn^.wCtrlID));

// sets output D0 to 0 [V]
if (wCtrlID = 100) then begin
   MessageBeep($FFFFFFFF);
   end;

// sets output D0 to more than 2 [V]
if (wCtrlID = 101) then begin
   MessageBeep(MB_ICONASTERISK);
   end;

// sets output D1 to 0 [V]
if (wCtrlID = 110) then begin
   MessageBeep(MB_ICONEXCLAMATION);
   end;

// sets output D1 to more than 2 [V]
if (wCtrlID = 111) then begin
   MessageBeep(MB_ICONHAND);
   end;

// sets Led OFF on the PCS10
if (wCtrlID = 300) then begin
   MessageBeep(MB_ICONQUESTION);
end;

// sets Led ON on the PCS10
if (wCtrlID = 301) then begin
   MessageBeep(MB_OK);
   end;

// sets the gain to 1 on all channels on the PCS10
if (wCtrlID = 401) then begin
   MessageBeep(0);
   end;

// sets the gain to 2 on all channels on the PCS10
if (wCtrlID = 402) then begin
   MessageBeep(0);
   end;

// sets the gain to 5 on all channels on the PCS10
if (wCtrlID = 405) then begin
   PlaySound('Self_defined_sound_1.wav', 0, SND_ASYNC or SND_FILENAME);
   end;

// sets the gain to 10 on all channels on the PCS10
if (wCtrlID = 410) then begin
   PlaySound('Self_defined_sound_2.wav', 0, SND_ASYNC or SND_FILENAME);
   end;

// Read voltage from the PCS10
if (wCtrlID = 700) then begin
   end;

// Read voltage from the PCS10
if (wCtrlID = 710) then begin
   end;

// sets output D0 to more than 2 [V]
if (wCtrlID = 801) then begin
   end;

// Did we click on the quit button ?
// Close the main window, thus quit the application.
if (wCtrlID = 999) then  PostMessage(poMyMainWin^.hWin, WM_CLOSE, 0, 0);

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyDialog.WMCommand

//******************************************************************
// TMyWindow = object(TWindow)
//******************************************************************

constructor TMyWindow.Init(strWinTitlePrm:ShortString);
//=====================================================
begin
inherited Init(strWinTitlePrm);
end;  // TMyWindow.Init

destructor TMyWindow.Done;
//========================
begin
inherited Done;
end;

procedure TMyWindow.WinCreate(strWinClass:ShortString);
//=====================================================
//var strName : string[50];
Begin
// We don't want to see the main window
//wSizeState := SIZE_MINIMIZED;

inherited WinCreate(strWinClass);

// Here calling :  DialogExecMSG(11199, @Self);  doesn't work,
// But posting a message to open the dialog box at startup is possible.
PostMessage(hWin, WM_COMMAND, 101, 0);

//strName:='zlportio' + char(0);
//DBW(0, 'result= ' + ITS(driverstart(@strName[1])));

end;  // TMyWindow.WinCreate

function TMyWindow.WMCommand(wPar:WPARAM; lPar:LPARAM):LRESULT;
//=============================================================
// treats the menu commands
begin
// Execute the dialog box at starting
// And the help window
if (lPar = 0) and (wPar =101) then DialogCreateMSG(11199, @Self);
//if (lPar = 0) and (wPar =121) then DialogCreateMSG(21199, @Self);

// Accelerators act as button click.
// Send the command to the active dialog
case LoWord(wPar) of
  999 : if (gglpoActiveDialog <> NIL) then
               gglpoActiveDialog^.WMCommand(LoWord(wPar), 0);
  end; // case

WMCommand:=inherited WMCommand(wPar, lPar);
end;  // TMyWindow.WMCommand

function TMyWindow.WindowProc(hWindow:HWND; wMsg:UINT; wPar:WPARAM; lPar:LPARAM):LRESULT;
//=======================================================================================
// All window messages are send here.

Begin
WindowProc:=inherited WindowProc(hWindow, wMsg, wPar, lPar);

{ if (wMsg <> WM_MouseMove)   and (wMsg <> WM_NCMouseMove) and
   (wMsg <> WM_MENUSELECT )   and (wMsg <> WM_ENTERIDLE ) and
   (wMsg <> WM_SetCursor)   and (wMsg <> WM_NCHittest) and
   (wMsg <> WM_SetFocus) and (wMsg <> WM_KillFocus) and
   (wMsg <> WM_KEYDOWN) and (wMsg <> WM_KEYUP) and
   (wMsg <> WM_Move)        and (wMsg <> WM_NCActivate) then
   DBW0(32, DBMsgToStr(wMsg) + ITS(wPar) + ITS(lPar)); }

//if (wMsg = WM_Destroy) then begin
//   end;

End;  // TMyWindow.WindowProc

//**********************************************************************
Begin // MAIN ******
// Assign the pointer on the function which will create the dialog.
// It is used by the library. So the library will construct
// a user defined object.
gglfnCreateDialogObj:=@myCreateDialogObj;

// Name under which the main window is registered. Spaces are not allowed
oApp.Init('Example072');

// Creation of the main window
new(poMyMainWin, Init('tEx072_Sound_Tests.pp'));

// The App.WindowClass can be modified here, if wanted. (Not usual)
oApp.Run;
oApp.Done;  // The main window object is disposed here.
End.

